from instaui.internal import import_presets
from instaui.internal.ast.core import WebWatchTask
from instaui.internal.ast import expression
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.exception import CodegenError
from instaui.internal.codegen.expr_codegen import ExpressionCodegen


class WebWatchTaskRenderer:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.import_table = ctx.imports
        self.bindings = ctx.root.binding_registry
        self.expr = ExpressionCodegen(ctx)

    def emit(self, tasks: list[WebWatchTask]) -> str:
        method = self.import_table.use_from_preset(
            import_presets.Instaui.web_watch_task_scheduler()
        )

        args_list = expression.ListExpr.from_values(
            *[self._parse_web_watch_task(task) for task in tasks]
        )
        code = f"{method}({self.expr.emit(args_list)})"
        return code

    def _parse_web_watch_task(self, task: WebWatchTask):
        try:
            key = self.bindings.get_value(task.fn.ref_id)
            return expression.ObjectExpr(
                props=[
                    *task.args.props,
                    expression.ObjectProperty(
                        key="hKey", value=expression.Literal(value=key)
                    ),
                ]
            )
        except Exception as e:
            raise CodegenError(f"Failed to generate code for {task}: {e}", task.source)



from __future__ import annotations
from enum import Enum, IntEnum
from pathlib import Path
from typing import Iterable, Optional, Union, Hashable
from instaui.systems.dataclass_system import dataclass, field
from itertools import chain


class CssRole(str, Enum):
    BASE = "base"  # plugin/reset/default
    COMPONENT = "component"  # regular component (default)
    THEME = "theme"  # theme
    OVERRIDE = "override"  # user override


class _CssLayer(IntEnum):
    BASE = 10
    COMPONENT = 20
    THEME = 30
    OVERRIDE = 40


ROLE_TO_LAYER: dict[CssRole, _CssLayer] = {
    CssRole.BASE: _CssLayer.BASE,
    CssRole.COMPONENT: _CssLayer.COMPONENT,
    CssRole.THEME: _CssLayer.THEME,
    CssRole.OVERRIDE: _CssLayer.OVERRIDE,
}


@dataclass(frozen=True)
class CssAsset:
    path_or_url: Union[Path, str]
    role: CssRole = CssRole.COMPONENT
    namespace: Optional[Hashable] = None

    @property
    def layer(self) -> _CssLayer:
        return ROLE_TO_LAYER[self.role]

    @property
    def is_local(self) -> bool:
        return isinstance(self.path_or_url, Path)

    def __hash__(self) -> int:
        """
        Anonymous assets need deduplication based on:
        - Same path_or_url
        - Same role
        - Same namespace
        """
        return hash((self.path_or_url, self.role, self.namespace))


@dataclass
class CssAssetCollection:
    by_namespace: dict[Hashable, CssAsset] = field(default_factory=dict)
    anonymous: list[CssAsset] = field(default_factory=list)

    def add(
        self,
        css: Union[Path, str, CssAsset],
        *,
        role: CssRole = CssRole.COMPONENT,
        namespace: Optional[Hashable] = None,
    ):
        if isinstance(css, CssAsset):
            asset = css
        else:
            asset = CssAsset(
                path_or_url=css,
                role=role,
                namespace=namespace,
            )

        if asset.namespace is not None:
            # namespace：override
            self.by_namespace[asset.namespace] = asset
        else:
            # anonymous：duplicate removal + order preservation
            if asset not in self.anonymous:
                self.anonymous.append(asset)

    def remove(
        self,
        css: Union[Path, str, CssAsset],
        *,
        role: CssRole = CssRole.COMPONENT,
        namespace: Optional[Hashable] = None,
    ) -> None:
        if isinstance(css, CssAsset):
            asset = css
        else:
            asset = CssAsset(
                path_or_url=css,
                role=role,
                namespace=namespace,
            )

        if asset.namespace is not None:
            self.by_namespace.pop(asset.namespace, None)
        else:
            try:
                self.anonymous.remove(asset)
            except ValueError:
                pass

    def iter_assets(self) -> Iterable[CssAsset]:
        """
        Returns CssAssets for HTML rendering (sorted by layer)
        """
        assets = chain(self.anonymous, self.by_namespace.values())
        assets = sorted(assets, key=lambda a: a.layer)
        return assets

    def iter_css_links(self) -> Iterable[Union[Path, str]]:
        for asset in self.iter_assets():
            yield asset.path_or_url

    @classmethod
    def merge(
        cls,
        global_assets: CssAssetCollection,
        local_assets: CssAssetCollection,
    ) -> CssAssetCollection:
        """
        Merge semantics:
        - Global assets added first
        - Local assets added after
        - Namespace automatically overridden by local
        - Order determined solely by role/layer
        """
        result = cls()

        for asset in global_assets.iter_assets():
            result.add(asset)

        for asset in local_assets.iter_assets():
            result.add(asset)

        return result



from instaui.internal.pipeline.assets import (
    AssetsPipeline,
    PluginCssProcessor,
    PluginImportMapProcessor,
    ComponentCssProcessor,
    ComponentImportMapProcessor,
)


WEB_ASSETS_PIPELINE = AssetsPipeline(
    [
        PluginCssProcessor(),
        PluginImportMapProcessor(),
        ComponentCssProcessor(),
        ComponentImportMapProcessor(),
    ]
)



from typing import Any

from instaui.internal.ui.bindable import (
    BindableMixin,
    is_bindable,
    mark_provided,
    mark_used,
)
from instaui.internal.ui.variable import Variable


class VariableReferenceName(BindableMixin):
    def __init__(self, variable: BindableMixin):
        self.variable = variable

    @property
    def _used(self) -> bool:
        return self.variable._used

    def _mark_used(self) -> None:
        return self.variable._mark_used()

    def _mark_provided(self) -> None:
        return self.variable._mark_provided()


def convert_reference(binding: Any):
    """
    Allows reactive variable configurations to be passed into custom components, enabling them to obtain their Ref references.

    Args:
        binding (Any): The reactive binding or variable to be converted

    Examples:
    .. code-block:: python

        class CustomElement(custom.element, esm="./custom_element.js"):
            def __init__(self, ref):
                super().__init__()
                self.props({"ref_binding": custom.convert_reference(ref)})
    """
    assert isinstance(binding, Variable) and is_bindable(binding), (
        "binding should be a Variable"
    )
    mark_used(binding)
    mark_provided(binding)

    return VariableReferenceName(binding)



from __future__ import annotations
from typing import Optional, Literal
from typing_extensions import Unpack
from instaui.internal.ui.element import Element
from .base_props import TLayoutBaseProps
from instaui.internal.ui.components._responsive_type._common import TMaybeResponsive


class Box(Element):
    def __init__(
        self,
        *,
        as_: Optional[Literal["div", "span"]] = None,
        as_child: Optional[bool] = None,
        display: Optional[
            TMaybeResponsive[Literal["none", "inline-block", "block", "contents"]]
        ] = None,
        **kwargs: Unpack[TLayoutBaseProps],
    ):
        """
        Creates a flexible box container with responsive layout properties.

        Args:
            as_ (Optional[Literal["div", "span"]]): HTML element type to render as.
            as_child (Optional[bool]): Whether to merge props with parent component.
            display (Optional[TMaybeResponsive[Literal["none", "inline-block", "block", "contents"]]]):
                CSS display property for the box.
            **kwargs (Unpack[TLayoutBaseProps]): Additional layout properties like padding, margin, etc.

        Example:
        .. code-block:: python
            # Basic box with padding
            with ui.box(p="1"):
                ui.text("Content")

            # Box with as_child to merge with parent
            with ui.box(p="1", as_child=True):
                ui.text("Merged content")

            # Responsive box with different padding per breakpoint
            with ui.box(
                p={"initial": "1px", "sm": "2", "md": "3", "lg": "4", "xl": "5"}
            ):
                ui.text("Responsive content")

            # Box with reactive padding
            padding = ui.state("1")
            with ui.box(p=padding):
                ui.text("Dynamic padding")
        """
        super().__init__("box")

        self.props(
            {
                "as": as_,
                "as_child": as_child,
                "display": display,
                **kwargs,
            }
        )



from __future__ import annotations
from typing import (
    Generic,
    Optional,
    Union,
    TypeVar,
)
from instaui.internal.ui.element import Element
from instaui.internal.ui.web_computed import WebComputed
from instaui.internal.ui.bindable import mark_used, is_bindable
from instaui.types import TVModelModifier

_T = TypeVar("_T")


class ValueElement(Element, Generic[_T]):
    def __init__(
        self,
        tag: Optional[str] = None,
        value: Union[_T, None] = None,
        is_html_component: bool = False,
        value_name: str = "value",
    ):
        super().__init__(tag)
        self.__is_html_component = is_html_component

        if value is not None:
            if is_bindable(value):
                if isinstance(value, WebComputed):
                    self.props({value_name: value})
                else:
                    self.vmodel(value, prop_name=value_name)
            else:
                self.props({value_name: value})

    def vmodel(
        self,
        value,
        modifiers: Union[TVModelModifier, list[TVModelModifier], None] = None,
        *,
        prop_name: str = "value",
    ):
        mark_used(value)
        return super().vmodel(
            value,
            modifiers,
            prop_name=prop_name,
            is_html_component=self.__is_html_component,
        )



from typing import Optional
from instaui.internal.backend.binding import (
    BackendBindingRegistryBase,
    UnboundBackendBindingRegistry,
)


class RuntimeContext:
    def __init__(
        self, *, binding_registry: Optional[BackendBindingRegistryBase] = None
    ):
        self.binding_registry: BackendBindingRegistryBase = (
            binding_registry or UnboundBackendBindingRegistry()
        )



from __future__ import annotations
from typing import Literal, Union
from typing_extensions import Unpack
from instaui.internal.ui.element import Element
from .base_props import TLayoutBaseProps
from instaui.internal.ui.components._responsive_type._common import (
    TMaybeResponsive,
    TLevel_0_9,
)


class TFlexBaseProps(TLayoutBaseProps, total=False):
    as_: Literal["div", "span"]
    as_child: bool
    display: TMaybeResponsive[Literal["none", "inline-flex", "flex"]]
    align: TMaybeResponsive[Literal["start", "center", "end", "stretch", "baseline"]]
    justify: TMaybeResponsive[Literal["start", "center", "end", "between"]]
    wrap: TMaybeResponsive[Literal["nowrap", "wrap", "wrap-reverse"]]
    gap: TMaybeResponsive[Union[str, TLevel_0_9]]
    gap_x: TMaybeResponsive[Union[str, TLevel_0_9]]
    gap_y: TMaybeResponsive[Union[str, TLevel_0_9]]


class TFlexProps(TFlexBaseProps, total=False):
    direction: TMaybeResponsive[
        Literal["row", "column", "row-reverse", "column-reverse"]
    ]


class Flex(Element):
    def __init__(
        self,
        **kwargs: Unpack[TFlexProps],
    ):
        """
        Creates a flex container with configurable layout and alignment.

        Args:
            **kwargs (Unpack[TFlexProps]): Flex container properties including:
                direction (TMaybeResponsive[Literal["row", "column", "row-reverse", "column-reverse"]]):
                    Flex direction for child elements.
                align (TMaybeResponsive[Literal["start", "center", "end", "stretch", "baseline"]]):
                    Cross-axis alignment of flex items.
                justify (TMaybeResponsive[Literal["start", "center", "end", "between"]]):
                    Main-axis alignment of flex items.
                wrap (TMaybeResponsive[Literal["nowrap", "wrap", "wrap-reverse"]]):
                    Flex wrapping behavior.
                gap (TMaybeResponsive[Union[str, TLevel_0_9]]): Gap between flex items.
                gap_x (TMaybeResponsive[Union[str, TLevel_0_9]]): Horizontal gap between items.
                gap_y (TMaybeResponsive[Union[str, TLevel_0_9]]): Vertical gap between items.
                display (TMaybeResponsive[Literal["none", "inline-flex", "flex"]]): Display type.
                as_ (TMaybeRef[Literal["div", "span"]]): HTML element type.
                as_child (TMaybeRef[bool]): Whether to merge with parent component.

        Example:
        .. code-block:: python
            # Basic flex container
            with ui.flex():
                ui.text("Item 1")
                ui.text("Item 2")

            # Row flex container
            with ui.row():
                ui.text("Horizontal items")

            # Column flex container
            with ui.column():
                ui.text("Vertical items")

            # Flex with gap and alignment
            with ui.flex(gap="2", align="center", justify="between"):
                ui.text("Left")
                ui.text("Center")
                ui.text("Right")
        """
        super().__init__("flex")
        self.props(kwargs)  # type: ignore


class FlexRow(Flex):
    def __init__(
        self,
        **kwargs: Unpack[TFlexBaseProps],
    ):
        """
        Creates a horizontal flex container with row direction.

        Args:
            **kwargs (Unpack[TFlexBaseProps]): Flex container properties including:
                align (TMaybeResponsive[Literal["start", "center", "end", "stretch", "baseline"]]):
                    Vertical alignment of flex items.
                justify (TMaybeResponsive[Literal["start", "center", "end", "between"]]):
                    Horizontal alignment of flex items.
                wrap (TMaybeResponsive[Literal["nowrap", "wrap", "wrap-reverse"]]):
                    Flex wrapping behavior.
                gap (TMaybeResponsive[Union[str, TLevel_0_9]]): Gap between flex items.
                gap_x (TMaybeResponsive[Union[str, TLevel_0_9]]): Horizontal gap between items.
                gap_y (TMaybeResponsive[Union[str, TLevel_0_9]]): Vertical gap between items.
                display (TMaybeResponsive[Literal["none", "inline-flex", "flex"]]): Display type.
                as_ (TMaybeRef[Literal["div", "span"]]): HTML element type.
                as_child (TMaybeRef[bool]): Whether to merge with parent component.

        Example:
        .. code-block:: python
            # Basic horizontal flex container
            with ui.row():
                ui.text("Item 1")
                ui.text("Item 2")

            # Row with gap and alignment
            with ui.row(gap="2", align="center", justify="between"):
                ui.text("Left")
                ui.text("Center")
                ui.text("Right")
        """
        super().__init__(direction="row", **kwargs)


class FlexColumn(Flex):
    def __init__(
        self,
        **kwargs: Unpack[TFlexBaseProps],
    ):
        """
        Creates a vertical flex container with column direction.

        Args:
            **kwargs (Unpack[TFlexBaseProps]): Flex container properties including:
                align (TMaybeResponsive[Literal["start", "center", "end", "stretch", "baseline"]]):
                    Horizontal alignment of flex items.
                justify (TMaybeResponsive[Literal["start", "center", "end", "between"]]):
                    Vertical alignment of flex items.
                wrap (TMaybeResponsive[Literal["nowrap", "wrap", "wrap-reverse"]]):
                    Flex wrapping behavior.
                gap (TMaybeResponsive[Union[str, TLevel_0_9]]): Gap between flex items.
                gap_x (TMaybeResponsive[Union[str, TLevel_0_9]]): Horizontal gap between items.
                gap_y (TMaybeResponsive[Union[str, TLevel_0_9]]): Vertical gap between items.
                display (TMaybeResponsive[Literal["none", "inline-flex", "flex"]]): Display type.
                as_ (TMaybeRef[Literal["div", "span"]]): HTML element type.
                as_child (TMaybeRef[bool]): Whether to merge with parent component.

        Example:
        .. code-block:: python
            # Basic vertical flex container
            with ui.column():
                ui.text("Item 1")
                ui.text("Item 2")

            # Column with gap and alignment
            with ui.column(gap="2", align="center", justify="between"):
                ui.text("Top")
                ui.text("Middle")
                ui.text("Bottom")
        """
        super().__init__(direction="column", **kwargs)



from __future__ import annotations
from typing import Any, Callable, Optional

from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui._expr.maybe_ref_expr import UIMaybeRefExpr
from instaui.internal.ui.imports import import_from
from instaui.internal.ui.protocol import CanOutputProtocol, ObservableProtocol
from instaui.protocol.ui.state import StateReadable
from instaui.systems.dataclass_system import dataclass
from instaui.debug.source import get_source_span
from .app_index import new_scope_if_needed
from .variable import Variable
from .enums import InputBindingType, OutputSetType
from .path_var import PathVar
from .bindable import BindableMixin, VarableBindHelper
from .app_context import get_current_scope


@dataclass(frozen=True)
class CustomRefMethod:
    module_name: str
    method_name: str


class CustomVar(
    PathVar,
    Variable,
    ObservableProtocol,
    CanOutputProtocol,
    BindableMixin,
):
    def __init__(
        self,
        module: str,
        method: str,
        *,
        args: Optional[Any] = None,
        deep_compare: bool = False,
        on_mark_used_fn: Optional[Callable[..., Any]] = None,
    ) -> None:
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        self._define_scope = get_current_scope()

        def on_register_fn():
            import_from(module, [method])
            if on_mark_used_fn:
                on_mark_used_fn()

        self._bind_helper = VarableBindHelper(
            self,
            self._define_scope.register_custom_ref,
            define_scope=self._define_scope,
            on_register_fn=on_register_fn,
            lazy_mark_used=[args],
        )

        self._method = CustomRefMethod(module, method)
        self._deep_compare = deep_compare
        self._args = _build_args(args)

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id

    @staticmethod
    def maybe_ref(value: Any):
        return UIMaybeRefExpr(value)

    @staticmethod
    def literal(value: Any):
        return UILiteralExpr(value)


class StateReadableCustomVar(CustomVar, StateReadable):
    pass


def _build_args(args):
    match args:
        case tuple():
            return tuple(UILiteralExpr.try_parse(arg) for arg in args)

        case list():
            return UILiteralExpr.try_parse_list(args)

        case dict():
            return UILiteralExpr.try_parse_dict(args)
        case None:
            return None
        case _:
            return UILiteralExpr.try_parse(args)



from __future__ import annotations
from typing import TYPE_CHECKING, Generator


if TYPE_CHECKING:
    from instaui.internal.ui.layout import Layout


class LayoutContext:
    def __init__(self, layouts: list[Layout]) -> None:
        self._layouts = sorted(
            layouts,
            key=lambda v: v.priority,
            reverse=True,
        )
        self._gens: list[Generator[None, None, None]] = []

    def __enter__(self):
        for layout in self._layouts:
            gen = layout.create_generator()
            try:
                next(gen)
            except StopIteration:
                raise RuntimeError("@ui.layout must yield exactly once")
            self._gens.append(gen)

        return self

    def __exit__(self, *_):
        for gen in reversed(self._gens):
            try:
                next(gen)
            except StopIteration:
                continue
            else:
                raise RuntimeError("@ui.layout must yield exactly once")

        self._gens.clear()
        return False



from typing import TypeVar, cast
from instaui.constants.ui import STORAGE_REF_METHOD, INSTAUI_JS_MODULE_NAME
from instaui.debug.api_boundary import user_api
from instaui.internal.ui.custom_var import CustomVar

_T = TypeVar("_T")


@user_api
def local_storage(key: str, value: _T, deep_compare: bool = False) -> _T:
    """
    Creates a reactive state object synchronized with the browser's local storage.

    This function initializes a reactive value tied to a given key in local storage.
    The state persists across page reloads and retains its value between sessions
    on the same browser and device.

    Args:
        key (str): The local storage key to associate with the value.
        value (_T): The default value to use if no value exists in local storage.

    Returns:
        _T: A reactive value linked to the specified local storage key.

    Example:
    .. code-block:: python

        from instaui import ui, html

        @ui.page('/')
        def index():
            name = ui.local_storage("username", "")
            html.input(name)
    """

    if isinstance(value, CustomVar):
        return value

    return cast(
        _T,
        CustomVar(
            INSTAUI_JS_MODULE_NAME,
            STORAGE_REF_METHOD,
            args={"type": "local", "key": key, "value": value},
            deep_compare=deep_compare,
        ),
    )



from typing import Callable, Hashable, Optional, Sequence
from instaui.systems import func_system


def create_handler_key(
    page_path: str,
    handler: Callable,
    debug_mode: bool,
    extra_key: Optional[Sequence[Hashable]] = None,
):
    _, lineno, _ = func_system.get_function_location_info(handler)
    key = f"path:{page_path}|line:{lineno}" if debug_mode else f"{page_path}|{lineno}"
    if extra_key:
        key = repr(tuple(extra_key) + (key,))

    return key



from __future__ import annotations
from pathlib import Path
from typing import Callable, Optional, Union
from instaui import __version__ as _INSTA_VERSION
from instaui import resources
from instaui.constants.runtime import RuntimeMode
from instaui.constants.ui import USER_ASSETS_ICONS_DIR_NAME
from instaui.internal.assets.context import GLOBAL_ASSETS, get_active_assets
from instaui.internal.assets.snapshot import AssetsSnapshot
from instaui.internal.backend.binding import BackendBindingRegistry
from instaui.internal.backend.fastapi.web_backend import FastapiBackend
from instaui.internal.codegen.program_codegen import ProgramCodegen
from instaui.internal.pipeline.ast_build.builder import AstBuilder
from instaui.internal.pipeline.ast_build.context import AstBuildInput
from instaui.internal.pipeline.web_html.html_renderer import (
    HtmlOptions,
    WebHtmlRenderer,
)
from instaui.internal.runtime.assets_init import inject_framework_assets
from instaui.internal.runtime.layout import LayoutContext
from instaui.internal.runtime.render_ctx import RenderContext
from instaui.internal.runtime.runtime_ctx import RuntimeContext
from instaui.internal.runtime.web.config import WebUserLaunchConfig
from instaui.internal.runtime.web.options import (
    WebUserRunOptions,
    WebUserRunWithOptions,
)
from instaui.internal.runtime.web.runtime_services import WebRuntimeServices
from instaui.internal.ui._app import App
from instaui.internal.ui.app_context import get_default_app
from instaui.internal.ui.app_index import new_app
from instaui.internal.router.registry import route_registry
from instaui.internal.pipeline.asset_resolver.backend_resolver import (
    BackendAssetResolver,
)
from instaui.internal.runtime.app_meta import AppMeta, ServerInfo, ServerRoute, AppIcons
from instaui.internal.pipeline.routing.fastapi import FastAPIRouteSyntaxAdapter

from .assets_pipeline import WEB_ASSETS_PIPELINE


class WebRuntime(WebRuntimeServices):
    def __init__(self, config: WebUserLaunchConfig):
        self.backend = FastapiBackend(
            config=config,
            runtimie_services=self,
            user_assets_dir=self.get_user_assets_folder(config.caller_folder_path),
        )
        self.asset_resolver = BackendAssetResolver(self.backend)
        self.route_syntax_adapter = FastAPIRouteSyntaxAdapter()
        self._launch_config = config
        self.binding_registry = BackendBindingRegistry(self.backend)
        self._mode = RuntimeMode.WEB
        self.debug = config.debug
        self._app_meta = AppMeta(
            self._mode,
            _INSTA_VERSION,
            config.debug,
            server_info=ServerInfo(
                watch_url=self.backend.endpoints.WATCH_URL_WITH_PREFIX,
                watch_async_url=self.backend.endpoints.ASYNC_WATCH_URL_WITH_PREFIX,
                event_url=self.backend.endpoints.EVENT_URL_WITH_PREFIX,
                event_async_url=self.backend.endpoints.ASYNC_EVENT_URL_WITH_PREFIX,
                download_url=self.backend.endpoints.DOWNLOAD_URL_WITH_PREFIX,
                assets_url=self.backend.endpoints.INSTAUI_USER_ASSETS_URL_WITH_PREFIX,
                assets_icons_name=USER_ASSETS_ICONS_DIR_NAME,
            ),
        )

        self.routes = route_registry.routes
        self.runtime_ctx = RuntimeContext(binding_registry=self.binding_registry)

        inject_framework_assets()

    def _render_page_core(self, route: str, caller: Callable) -> str:
        layout_ctx = LayoutContext(get_default_app().layouts)

        with new_app(self._mode, debug=self.debug) as app:
            with layout_ctx:
                caller(self.routes[route].fn)

            updated_app_meta = self._update_app_meta(app, route)
            page_assets = get_active_assets()
            assets_snapshot = WEB_ASSETS_PIPELINE.run(GLOBAL_ASSETS, page_assets)

        render_ctx = RenderContext(app_meta=updated_app_meta, route=route)

        ast = AstBuilder(self.runtime_ctx, render_ctx, self.asset_resolver).build(
            self._create_ast_build_input(app, assets_snapshot)
        )
        code = ProgramCodegen(self.runtime_ctx).emit(ast)
        renderer = WebHtmlRenderer()
        html_str = renderer.render(code, self._build_render_options(assets_snapshot))

        return html_str

    def run(self, optons: WebUserRunOptions):
        self.backend.run(optons)

    def run_with(self, options: WebUserRunWithOptions):
        self.backend.run_with(options)

    def _create_ast_build_input(
        self, app: App, assets_pool: AssetsSnapshot
    ) -> AstBuildInput:
        return AstBuildInput(app=app, scopes=app.created_scopes, assets=assets_pool)

    def _build_render_options(self, merged_assets: AssetsSnapshot) -> HtmlOptions:
        import_maps = (
            merged_assets.import_maps
            # | self._options.get_import_maps_cdn_overrides()
        )

        return HtmlOptions(
            is_debug=self.debug,
            prefix=self._launch_config.prefix,
            favicon_url=self._normalize_asset_path(
                merged_assets.favicon or resources.FAVICON_PATH
            ),
            css_links=[
                self._normalize_asset_path(css) for css in merged_assets.css_links
            ],
            js_links=[self._normalize_asset_path(js) for js in merged_assets.js_links],
            import_map={
                name: self._normalize_asset_path(link)
                for name, link in import_maps.items()
            },
            style_tags=[
                (style_tag.group_id, style_tag.content)
                for style_tag in merged_assets.style_tags
            ],
            script_tags=[
                (script_tag.script_attrs, script_tag.content)
                for script_tag in merged_assets.script_tags
            ],
        )

    def _normalize_asset_path(self, path: Union[str, Path, None]) -> str:
        if path is None:
            return ""
        if isinstance(path, str):
            return path

        return self.asset_resolver.resolve_asset(path)

    def render_fn(self, route: str) -> str:
        return self._render_page_core(route, Caller.call)

    async def render_fn_async(self, route: str) -> str:
        return self._render_page_core(route, Caller.call_async)

    def get_user_assets_folder(self, base_folder: Path) -> Optional[Path]:
        assets_folder = base_folder.joinpath("assets")
        if assets_folder.exists():
            return assets_folder

        return None

    def _update_app_meta(self, app: App, route: str):
        icons = AppIcons()
        if app._used_icons:
            icons.names = list(app._used_icons)
        if app._used_icons_set:
            icons.sets = list(app._used_icons_set)

        if not icons:
            icons = None

        return self._app_meta.replace(
            app_icons=icons,
            route=ServerRoute(self.route_syntax_adapter.normalize(route)),
        ).to_dict()


class Caller:
    @staticmethod
    def call(fn: Callable):
        return fn()

    @staticmethod
    async def call_async(fn: Callable):
        return await fn()



from __future__ import annotations
from contextlib import contextmanager
import contextvars
from pathlib import Path
from typing import TYPE_CHECKING, Any, Hashable, Optional, Union
from instaui.internal.assets.component_dep import ComponentDependencyInfo
from instaui.internal.assets.css_assets import CssAsset, CssRole
from instaui.internal.assets.plugin import PluginDependencyInfo, PluginOptionsBuilder
from instaui.internal.assets.script_assets import ScriptTag
from instaui.internal.assets.style_assets import StyleTag
from instaui.internal.context import in_page_context
from .base import AssetsDeclaration

if TYPE_CHECKING:
    from instaui.internal.ui.element import Element

GLOBAL_ASSETS = AssetsDeclaration()
LOCAL_ASSETS: contextvars.ContextVar[Optional[AssetsDeclaration]] = (
    contextvars.ContextVar("local_assets", default=None)
)


@contextmanager
def enter_assets_context():
    token = LOCAL_ASSETS.set(AssetsDeclaration())
    try:
        yield
    finally:
        LOCAL_ASSETS.reset(token)


def get_active_assets() -> AssetsDeclaration:
    """
    Return the AssetsPool that should actually be written based on context.
    """
    if in_page_context():
        local = LOCAL_ASSETS.get()
        if local is None:
            local = AssetsDeclaration()
            LOCAL_ASSETS.set(local)
        return local
    else:
        return GLOBAL_ASSETS


def add_js_link(url: Union[str, Path]):
    pool = get_active_assets()
    pool.js_links.add(url)


def remove_js_link(url: Union[str, Path]):
    pool = get_active_assets()
    pool.js_links.discard(url)


def add_css_link(
    css: Union[Path, str],
    *,
    role: Union[str, CssRole] = CssRole.COMPONENT,
    namespace: Optional[Hashable] = None,
):
    """Add a CSS link to the HTML document."""
    role = CssRole(role) if isinstance(role, str) else role
    get_active_assets().css_links.add(css, namespace=namespace, role=role)


def remove_css_link(
    css: Union[Path, str, CssAsset],
    *,
    role: Union[str, CssRole] = CssRole.COMPONENT,
    namespace: Optional[Hashable] = None,
):
    role = CssRole(role) if isinstance(role, str) else role
    get_active_assets().css_links.remove(css, namespace=namespace, role=role)


def use_favicon(path: Path):
    """Set the favicon of the HTML document.

    Args:
        favicon (Path): The path to the favicon.
    """
    get_active_assets().favicon = path


def add_import_map(name: str, link: Union[str, Path]):
    """Add an import map to the HTML document."""
    get_active_assets().import_maps[name] = link


def register_plugin(
    name: str,
    esm: Path,
    *,
    externals: Optional[dict[str, Path]] = None,
    css: Optional[list[Union[str, Path, CssAsset]]] = None,
    options: Optional[dict | PluginOptionsBuilder] = None,
):
    """
    Register a plugin

    Args:
        name (str): Plugin name
        esm (Path): Path to plugin's ESM file. Format should match vue.js plugins
        externals (Optional[dict[str, Path]], optional): External dependencies for plugin JS implementation. Defaults to None.
        css (Optional[list[Path]], optional): List of CSS file paths for the plugin. Defaults to None.
        options (Optional[dict | PluginOptionsBuilder], optional): Plugin options. Defaults to None.

    Example:
    .. code-block:: python
        from instaui import custom

        # Basic usage
        custom.register_plugin("my-plugin", Path(__file__).parent / "my-plugin.esm.js")

        # With external dependencies
        custom.register_plugin(
            "my-plugin",
            Path(__file__).parent / "my-plugin.esm.js",
            externals={"echarts": Path(__file__).parent / "echarts.esm.js"})

        # With install method options
        custom.register_plugin(..., options = {'opt1': 'value1', 'opt2': 'value2'})

        # With component-dependent options (only needed for complex plugins that define components, like vue-router)
        from instaui import custom

        def parse_scope(self, parse_key: StringKeyFn, parse_scope: ParseScopeFn):
            # parse_key ensures JS object keys are properly quoted
            # parse_scope translates Scope objects into component references
            return {parse_key(k): parse_scope(v) for k, v in self.routes.items()}

        custom.register_plugin(..., options = parse_scope)
    """
    info = PluginDependencyInfo(
        f"plugin/{name}", esm, externals or {}, css or [], options=options
    )
    get_active_assets().plugins.add(info)
    return info


def clear_registered_plugins():
    get_active_assets().plugins.clear()


def add_script_tag(content: str, *, script_attrs: Optional[dict[str, Any]] = None):
    """Add a script tag to the HTML document."""
    get_active_assets().script_tags.append(ScriptTag(content, script_attrs))


def add_js_code(code: str, *, script_attrs: Optional[dict[str, Any]] = None):
    """Add a script tag to the HTML document with the given JavaScript code.

    Args:
        code (str): The JavaScript code.
        script_attrs (Optional[dict[str, Any]], optional): The attributes of the script tag. Defaults to None.
    """
    add_script_tag(code, script_attrs=script_attrs)


def add_style_tag(content: str, *, group_id: Optional[str] = None):
    """Add a style tag to the HTML document.

    Args:
        content (str): The content of the style tag.
        group_id (Optional[str], optional): The group id of the style tag. Defaults to None.
    """
    get_active_assets().style_tags.append(StyleTag(content, group_id))


def add_component_dependency(element: type[Element], info: ComponentDependencyInfo):
    get_active_assets().component_dependencies.add(element, info)
    return info


def register_component_extension(
    *,
    target: type[Element],
    kind: str,
    values: list[str],
):
    """
    Invoked by third-party component authors to declare the additional dependencies required for a component family in Zero mode.
    """
    get_active_assets().component_extensions.add(
        target=target,
        kind=kind,
        values=values,
    )



from abc import ABC, abstractmethod
from pathlib import Path
from instaui.systems import file_system
from .resolver import AssetResolver


class AssetMaterializer(ABC):
    """
    Responsible for "materializing" source assets to the target location.
    AssetResolver is only responsible for calculating the target_path.
    Whether to create, copy, and how to copy is determined by this protocol.
    """

    @abstractmethod
    def ensure_dir(self, path: Path) -> None:
        """Ensure the directory exists (idempotent)"""
        pass

    @abstractmethod
    def copy_file(self, src: Path, dst: Path) -> None:
        """Copy a single file"""
        pass

    @abstractmethod
    def copy_dir(self, src: Path, dst: Path) -> None:
        """Copy an entire directory"""
        pass


class WebViewAssetResolver(AssetResolver):
    def __init__(
        self,
        assets_dir_path: Path,
        materializer: AssetMaterializer,
    ) -> None:
        self.assets_dir_path = assets_dir_path
        self.materializer = materializer

    def resolve_asset(self, path: Path) -> str:
        hash_part = file_system.generate_hash_name_from_path(path.parent)
        new_folder_path = self.assets_dir_path.joinpath(hash_part)
        new_path = new_folder_path.joinpath(path.name)

        self.materializer.ensure_dir(new_folder_path)

        if path.is_file():
            self.materializer.copy_file(path, new_path)
            return "./" + self._convert_to_relative(new_path)
        else:
            self.materializer.copy_dir(path, new_path)
            return "./" + self._convert_to_relative(new_path) + "/"

    def _convert_to_relative(self, file_path: Path, relative_parent=False):
        return str(
            file_path.relative_to(
                self.assets_dir_path.parent if relative_parent else self.assets_dir_path
            )
        ).replace("\\", "/")



from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder
from instaui.internal.pipeline.ast_build.utils import ui_bindings_utils

from instaui.internal.ui._scope import Scope
from instaui.internal.ast import core as ast_core, expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui.js_watch import JsWatch

if TYPE_CHECKING:
    pass


class JsWatchParser:
    def __init__(
        self,
        builder_ctx: AstBuilderContext,
    ) -> None:
        self.ctx = builder_ctx
        self.expr = ExprBuilder(builder_ctx)

    def parse(self, scope: Scope) -> Optional[list[ast_core.JsWatchCall]]:
        if not scope._js_watchs:
            return None

        return [self._parse_js_watch(watch) for watch in scope._js_watchs]

    def _parse_js_watch(self, watch: JsWatch):
        outputs = list(watch._outputs or [])
        inputs = ui_bindings_utils.auto_made_inputs_to_slient(
            list(watch._inputs or []), outputs
        )
        inputs, slients, _ = ui_bindings_utils.analyze_observable_inputs(inputs)
        slients = ui_bindings_utils.normalize_int_values(slients)

        output_types = ui_bindings_utils.normalize_int_values(
            [x._to_event_output_type().value for x in outputs]
        )

        args = expression.ObjectExpr.from_kwargs(
            code=expression.Literal(watch.code),
            inputs=self.expr.list_expr(mode="literal", values=inputs)
            if inputs
            else None,
            outputs=self.expr.list_expr(mode="literal", values=outputs)
            if outputs
            else None,
            opTypes=expression.ListLiteral(list(output_types))
            if output_types
            else None,
            slient=expression.ListLiteral(list(slients)) if slients else None,
            immediate=None
            if watch._immediate
            else expression.Literal(watch._immediate),
            deep=None if watch._deep else expression.Literal(watch._deep),
            once=expression.Literal(watch._once) if watch._once else None,
            flush=expression.Literal(watch._flush) if watch._flush else None,
        )

        return ast_core.JsWatchCall(args)



from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from instaui.debug.model import SourceSpan


class CodegenError(Exception):
    def __init__(self, message: str, source: SourceSpan | None):
        super().__init__(message)
        self.message = message
        self.source = source

    def __str__(self) -> str:
        if self.source is not None:
            return f"{self.message}\n--> {self.source}"
        return self.message



from __future__ import annotations
from typing import (
    Any,
    Optional,
    Sequence,
    Union,
)

from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.bindable import mark_used
from . import watch_types


class ExprWatch:
    def __init__(
        self,
        sources: Union[Any, Sequence],
        code: str,
        *,
        bindings: Optional[dict[str, Any]] = None,
        immediate: bool = False,
        deep: Union[bool, int] = False,
        once: bool = False,
        flush: Optional[watch_types.TFlush] = None,
    ) -> None:
        mark_used(sources)

        self.code = code

        if not isinstance(sources, Sequence):
            sources = [sources]

        self._on = UILiteralExpr.try_parse_list(sources)
        self._bindings = UILiteralExpr.try_parse_dict(bindings) if bindings else None

        self._immediate = immediate
        self._deep = deep
        self._once = once
        self._flush = flush


def expr_watch(
    sources: Union[Any, Sequence],
    code: str,
    *,
    bindings: Optional[dict[str, Any]] = None,
    immediate: bool = False,
    deep: Union[bool, int] = False,
    once: bool = False,
    flush: Optional[watch_types.TFlush] = None,
):
    """ """

    watch = ExprWatch(
        sources,
        code,
        bindings=bindings,
        immediate=immediate,
        deep=deep,
        once=once,
        flush=flush,
    )

    get_current_scope().register_expr_watch(watch)
    return watch



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Optional

from instaui.internal.ast.core import App
from instaui.internal.codegen.components.codegen_components import CodegenComponents
from instaui.internal.codegen.context import FileCodegenContext, CodegenRootContext
from instaui.systems.dataclass_system import dataclass
from .file_codegen import FileCodegen
from .file_unit import FileUnit
from .component_file_map import ComponentFileMap

if TYPE_CHECKING:
    from instaui.internal.runtime.runtime_ctx import RuntimeContext


@dataclass()
class CodegenResult:
    js: dict[str, str]
    entry_name: str
    app_var: str
    css: Optional[str] = None

    @property
    def entry_code(self) -> str:
        return self.js[self.entry_name]


class ProgramCodegen:
    def __init__(
        self,
        runtime_ctx: RuntimeContext,
        components: Optional[CodegenComponents] = None,
    ) -> None:
        self.root_ctx = CodegenRootContext(
            binding_registry=runtime_ctx.binding_registry,
            components=components,
        )

    def emit(
        self, app: App, *, mode: Literal["single", "multi"] = "single"
    ) -> CodegenResult:
        """
        return: {path: code}
        """

        if mode == "single":
            return self._emit_single(app)
        if mode == "multi":
            return self._emit_multi(app)

        raise ValueError(mode)

    def _emit_single(self, app: App) -> CodegenResult:
        from .app import AppCodegen

        file_ctx = FileCodegenContext(self.root_ctx)
        code = AppCodegen(file_ctx).emit(app)

        app_var = file_ctx.names.resolve(app.bootstrap.target)
        return CodegenResult({"main.js": code}, entry_name="main.js", app_var=app_var)

    def _emit_multi(self, app: App) -> CodegenResult:
        files = self._split_to_files(app)

        file_map = ComponentFileMap()
        for unit in files:
            for comp in unit.components:
                file_map.register(comp.id, unit.path)

        result = {}
        for unit in files:
            code = FileCodegen(file_map, self.root_ctx).emit(unit, app)
            result[unit.path] = code

        return CodegenResult(result, "main.js", "app")

    def _split_to_files(self, app: App) -> list[FileUnit]:
        units: list[FileUnit] = []

        for comp in app.components[1:]:
            units.append(
                FileUnit(
                    path=f"components/{comp.id.uid}.js",
                    components=[comp],
                )
            )

        units.append(
            FileUnit(
                path="main.js",
                components=[app.components[0]],
                is_entry=True,
            )
        )

        return units



from contextlib import ExitStack, contextmanager
from typing import Optional

from instaui.constants.runtime import RuntimeMode
from instaui.internal.assets import enter_assets_context
from instaui.internal.context.page_context import enter_page_context

from ._app import App, DefaultApp
from ._scope import Scope
from .app_context import app_var, get_app


@contextmanager
def new_app(
    mode: RuntimeMode,
    *,
    debug: bool,
    meta: Optional[dict] = None,
):
    app = App(
        mode=mode,
        debug=debug,
        meta=meta,
    )
    token = app_var.set(app)
    app_scope = Scope()

    with ExitStack() as stack:
        stack.callback(app_var.reset, token)
        stack.enter_context(enter_page_context())
        stack.enter_context(enter_assets_context())
        stack.enter_context(app_scope)
        yield app


def check_default_app_slot_or_error(
    error_message="Operations are not allowed outside of ui.page",
):
    if isinstance(get_app(), DefaultApp):
        raise ValueError(error_message)


def new_scope_if_needed() -> None:
    ctx = get_app()

    if ctx.has_pending_scope():
        pending = ctx.top_pending_scope()

        if not pending.realized:
            container = ctx.get_current_container()
            scope = Scope()
            scope.__enter__()
            container.add_child(scope)
            container._bind_scope(scope)
            pending.realized = True



from abc import ABC, abstractmethod
import typing

from .event_modifier import TEventModifier


class EventMixin(ABC):
    @abstractmethod
    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]] = None,
    ):
        pass



from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
from instaui.internal.assets.snapshot import AssetsSnapshot
from instaui.internal.pipeline.ast_build.component_scope import ComponentScope
from instaui.internal.pipeline.ast_build.symbol_mapper import IdGenerator, SymbolMapper
from instaui.internal.ui._app import App
from instaui.internal.ui._scope import Scope
from instaui.systems.dataclass_system import dataclass

if TYPE_CHECKING:
    from instaui.internal.backend.binding import BackendBindingRegistryBase
    from instaui.internal.runtime.render_ctx import RenderContext
    from instaui.internal.pipeline.asset_resolver.resolver import AssetResolver


class AstBuilderContext:
    def __init__(
        self,
        binding_registry: BackendBindingRegistryBase,
        render_ctx: RenderContext,
        asset_resolver: AssetResolver,
    ) -> None:
        self._shared_id_generator = IdGenerator()
        self.scope_mapper = SymbolMapper(self._shared_id_generator)
        self.var_mapper = SymbolMapper(self._shared_id_generator)
        self.const_var_mapper = SymbolMapper(self._shared_id_generator)
        self.binding_registry = binding_registry
        self._app_ref = object()
        self.render_ctx = render_ctx
        self.asset_resolver = asset_resolver
        # Component Scope Management
        self.component_scope_manager = ComponentScope()

    def get_app_id(self):
        return self.const_var_mapper.get_or_create_id(self._app_ref)


@dataclass(frozen=True)
class AstBuildInput:
    app: App
    scopes: Sequence[Scope]
    assets: AssetsSnapshot



from __future__ import annotations
from typing import (
    Generator,
    Literal,
    Optional,
    Tuple,
    Union,
    Sequence,
    Generic,
    TypeVar,
    cast,
)
from contextlib import contextmanager
from enum import Enum
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.pending_scope import PendingScope
from instaui.systems.dataclass_system import dataclass
from instaui.internal.ui._scope import Scope
from instaui.internal.ui.app_context import get_app, get_current_container
from instaui.internal.ui.renderable import Renderable, ScopeBoundaryRenderable
from instaui.internal.ui.bindable import is_bindable, mark_used
from instaui.internal.ui.container import Container, ScopeGuard
from instaui.internal.ui.vfor_item import VForItem, VForIndex, VForItemKey

_T = TypeVar("_T")


@dataclass()
class _VForRangeInfo:
    end: int
    start: int = 1
    step: int = 1

    def mark_used(self):
        mark_used(self.end)
        mark_used(self.start)
        mark_used(self.step)

    def to_dict(self):
        return {
            "end": self.end,
            "start": self.start if self.start != 1 else None,
            "step": self.step if self.step != 1 else None,
        }


class VForArrayTypeEnum(Enum):
    CONST = "c"
    REF = "r"
    RANGE = "n"


class VFor(ScopeBoundaryRenderable, Container, Generic[_T]):
    def __init__(
        self,
        data: Union[Sequence[_T], dict[str, _T], _T, None],
        *,
        key: Union[Literal["item", "index"], str] = "index",
        range: Optional[_VForRangeInfo] = None,
    ):
        """for loop component.

        Args:
            data (Union[Sequence[_T], ElementBindingMixin[list[_T]]]): data source.
            key (Union[Literal[&quot;item&quot;, &quot;index&quot;], str]]): key for each item. Defaults to 'index'.

        Examples:
        .. code-block:: python
            items = ui.state([1,2,3])

            with ui.vfor(items) as item:
                html.span(item)

            # object key
            items = ui.state([{"name": "Alice"}, {"name": "Bob"}])
            with ui.vfor(items, key="item.name") as item:
                html.span(item.name)

            # js computed key
            items = ui.state([{"name": "Alice"}, {"name": "Bob"}])
            with ui.vfor(items, key=": (item , index) => item.name + index") as item:
                html.span(item.name)

            # iter info
            items = ui.state({"a": 1, "b": 2, "c": 3})
            with ui.vfor(items).with_key() as [value, key]:

                html.span(key)
                html.span(value)

            # range
            with ui.vfor.range(10) as i:
                html.paragraph(i)
        """

        super().__init__()
        get_current_container().add_child(self)

        if is_bindable(data):
            self._array_type = VForArrayTypeEnum.REF
        else:
            self._array_type = VForArrayTypeEnum.CONST

        if range is not None:
            assert data is None, "range and data cannot be used together"
            self._array_type = VForArrayTypeEnum.RANGE
            mark_used(range.to_dict())

        mark_used(data)
        self._data = UILiteralExpr.try_parse(data) if data is not None else None
        self._range = range.to_dict() if range else None
        self._key = key
        self._transition_group_setting = None

        self._vfor_item = VForItem()
        self._vfor_index = VForIndex()
        self._vfor_item_key = VForItemKey()
        self._renderables: list[Renderable] = []
        self._scope_guard = ScopeGuard()

    def __enter__(self) -> _T:
        super().__enter__()
        get_app().push_pending_scope(PendingScope())
        return cast(_T, self._vfor_item)

    def __exit__(self, *_):
        get_app().pop_pending_scope()
        super().__exit__(*_)

    @contextmanager
    def with_index(self) -> Generator[Tuple[_T, int], None, None]:
        with self:
            yield (
                self._vfor_item,
                self._vfor_index,
            )  # type: ignore

    @contextmanager
    def with_key(self) -> Generator[Tuple[_T, str], None, None]:
        with self:
            yield (self._vfor_item, self._vfor_item_key)  # type: ignore

    @contextmanager
    def with_key_index(self) -> Generator[Tuple[_T, str, int], None, None]:
        with self:
            yield (
                self._vfor_item,
                self._vfor_item_key,
                self._vfor_index,
            )  # type: ignore

    def _set_range_type(self):
        self._array_type = VForArrayTypeEnum.RANGE

    def transition_group(self, name="fade", tag: Optional[str] = None):
        self._transition_group_setting = {"name": name, "tag": tag}
        return self

    @classmethod
    def range(
        cls,
        end: int,
        *,
        start: int = 0,
        step: int = 1,
    ) -> VFor[int]:
        obj = cls(None, range=_VForRangeInfo(start=start, end=end, step=step))

        return obj  # type: ignore

    def add_child(self, renderable: Renderable):
        self._renderables.append(renderable)

    def _bind_scope(self, scope: Scope):
        self._scope_guard.bind_scope(scope)

    def _release_scope(self):
        self._scope_guard.release_scope()



import typing
from instaui.debug.api_boundary import user_api
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_scope
from instaui.internal.ui.app_index import new_scope_if_needed
from instaui.internal.ui.bindable import BindableMixin, VarableBindHelper, mark_used
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier
from instaui.internal.ui.variable import Variable
from instaui.debug.source import get_source_span


class ExprEvent(Variable, EventMixin, BindableMixin):
    def __init__(
        self,
        *,
        code: str,
        bindings: typing.Optional[dict[str, typing.Any]] = None,
    ):
        self._source_span_ = get_source_span()
        new_scope_if_needed()
        define_scope = get_current_scope()
        self._code = code
        self._bindings = UILiteralExpr.try_parse_dict(bindings or {})

        self._bind_helper = VarableBindHelper(
            self,
            define_scope.register_expr_event,
            define_scope=define_scope,
            lazy_mark_used=[self._bindings],
        )

    def _attach_to_element(
        self,
        *,
        extends: typing.Optional[typing.Sequence],
        modifier: typing.Optional[typing.Sequence[TEventModifier]],
    ):
        mark_used(extends)
        mark_used(modifier)
        self._mark_used()

        return self

    @property
    def _used(self) -> bool:
        return self._bind_helper.used

    def _mark_used(self) -> None:
        self._bind_helper.mark_used()

    def _mark_provided(self) -> None:
        self._bind_helper.mark_provided()

    @property
    def _define_scope_id(self) -> int:
        return self._bind_helper._define_scope.id


@user_api
def expr_event(
    *,
    code: str,
    bindings: typing.Optional[dict[str, typing.Any]] = None,
):
    """
    Create an event object that can be bound to a UI component's event listener.

    This function generates a callable event handler with optional contextual bindings.
    The event logic is defined via a code string, which can reference bound variables.

    Args:
        code (str): A string containing the executable logic for the event handler.
                    Typically contains a function body or expression that utilizes bound variables.
        bindings (typing.Optional[dict[str, typing.Any]], optional): A dictionary mapping variable names to values that should be available in the
            event handler's context. If None, no additional bindings are created.. Defaults to None.

    Example:
    .. code-block:: python
        a = ui.state(1)

        event = ui.vue_event(bindings={"a": a}, code=r'''()=> { a.value +=1}''')

        html.span(a)
        html.button("plus").on("click", event)
    """
    return ExprEvent(code=code, bindings=bindings)



from typing import Any, Optional, Sequence
from fastapi import HTTPException, Response
from instaui.internal.backend.common.response import response_data
import pydantic


class ResponseData(pydantic.BaseModel):
    values: Optional[list[Any]] = None
    types: Optional[Sequence[int]] = None


def response_web_data(
    outputs_binding_count: int, result: Any, response: Optional[Response] = None
):
    new_result = result
    if response is not None:
        response_interceptor, new_result = _extract_response_interceptor(
            outputs_binding_count, result
        )
        if response_interceptor is not None:
            response_interceptor(response)

    return ResponseData(**response_data(outputs_binding_count, new_result))


def _extract_response_interceptor(outputs_binding_count: int, result: Any):
    if outputs_binding_count == 0:
        if callable(result):
            return result, None
        return None, result

    if outputs_binding_count == 1:
        if isinstance(result, tuple) and len(result) == 2 and callable(result[1]):
            return result[1], result[:-1]
        return None, result

    if (
        isinstance(result, tuple)
        and len(result) - outputs_binding_count == 1
        and callable(result[-1])
    ):
        return result[-1], result[:-1]
    return None, result


class HandlerNotFoundError(HTTPException):
    def __init__(
        self,
        message: str,
    ) -> None:
        super().__init__(404, detail=message)



from __future__ import annotations
from typing import TYPE_CHECKING

from instaui.internal.ui._scope import Scope
from .container import Container, ScopeGuard
from .slot_prop import SlotProp

if TYPE_CHECKING:
    from .renderable import Renderable


class Slot(Container):
    def __init__(self) -> None:
        super().__init__()
        self.children: list[Renderable] = []
        self._scope_guard = ScopeGuard()
        self._slot_prop = SlotProp()

    def slot_props(self, prop_name: str):
        return self._slot_prop.get(prop_name)

    def add_child(self, child: Renderable):
        self.children.append(child)

    def __iter__(self):
        return iter(self.children)

    def __len__(self):
        return len(self.children)

    def _bind_scope(self, scope: Scope):
        self._scope_guard.bind_scope(scope)

    def _release_scope(self):
        self._scope_guard.release_scope()



__all__ = [
    "MISSING",
    "PageState",
    "StateModel",
    "TVForItem",
    "add_css_link",
    "add_import_map",
    "add_js_code",
    "add_js_link",
    "add_script_tag",
    "add_style",
    "box",
    "column",
    "computed",
    "container",
    "content",
    "data",
    "element",
    "element_ref",
    "event",
    "event_context",
    "flex",
    "grid",
    "heading",
    "icon",
    "image",
    "js_computed",
    "js_event",
    "js_fn",
    "js_output",
    "js_watch",
    "layout",
    "lazy_render",
    "len_",
    "link",
    "local_storage",
    "match",
    "not_",
    "page",
    "route_param",
    "patch_set",
    "pre_setup_action",
    "query_param",
    "remove_css_link",
    "row",
    "run_element_method",
    "server",
    "session_storage",
    "skip_output",
    "slient",
    "state",
    "str_format",
    "to_value",
    "teleport",
    "text",
    "timer",
    "unwrap_reactive",
    "use_dark",
    "use_favicon",
    "use_language",
    "use_page_title",
    "vfor",
    "video",
    "vif",
    "expr_computed",
    "expr_event",
    "expr_watch",
    "watch",
    "webview",
    "__version__",
]


from instaui.internal.assets import (
    add_css_link,
    add_import_map,
    add_js_code,
    add_js_link,
    add_script_tag,
    remove_css_link,
    use_favicon,
)
from instaui.internal.assets import (
    add_style_tag as add_style,
)
from instaui.internal.router.api import page
from instaui.internal.ui._layout.box import Box as box
from instaui.internal.ui._layout.container import Container as container
from instaui.internal.ui._layout.flex import (
    Flex as flex,
)
from instaui.internal.ui._layout.flex import (
    FlexColumn as column,
)
from instaui.internal.ui._layout.flex import (
    FlexRow as row,
)
from instaui.internal.ui._layout.grid import Grid as grid
from instaui.internal.ui.components.content import Content as content
from instaui.internal.ui.components.heading import Heading as heading
from instaui.internal.ui.components.icon import Icon as icon
from instaui.internal.ui.components.lazy_render import LazyRender as lazy_render
from instaui.internal.ui.components.link import Link as link
from instaui.internal.ui.components.teleport import teleport
from instaui.internal.ui.components.text import Text as text
from instaui.internal.ui.components.timer.timer import Timer as timer
from instaui.internal.ui.const_data import const_data as data
from instaui.internal.ui.element import Element as element
from instaui.internal.ui.element_ref import (
    ElementRef as element_ref,
)
from instaui.internal.ui.element_ref import (
    run_element_method,
)
from instaui.internal.ui.event_context import EventContext as event_context
from instaui.internal.ui.expr import len_, not_
from instaui.internal.ui.expr_computed import ExprComputed as expr_computed
from instaui.internal.ui.expr_event import expr_event
from instaui.internal.ui.expr_watch import expr_watch
from instaui.internal.ui.html.image import Image as image
from instaui.internal.ui.html.video import Video as video
from instaui.internal.ui.input_slient_data import InputSilentData as slient
from instaui.internal.ui.js_computed import js_computed
from instaui.internal.ui.js_event import js_event
from instaui.internal.ui.js_fn import js_fn
from instaui.internal.ui.js_output import JsOutput as js_output
from instaui.internal.ui.js_watch import js_watch
from instaui.internal.ui.layout import layout
from instaui.internal.ui.local_storage import local_storage
from instaui.internal.ui.match import Match as match
from instaui.internal.ui.missing import MISSING
from instaui.internal.ui.page_state import PageState
from instaui.internal.ui.params import query_param, route_param
from instaui.internal.ui.patch_update import patch_set
from instaui.internal.ui.pre_setup import PreSetupAction as pre_setup_action
from instaui.internal.ui.session_storage import session_storage
from instaui.internal.ui.skip import skip_output
from instaui.internal.ui.state import StateModel, state
from instaui.internal.ui.str_format import str_format
from instaui.internal.ui.to_value import to_value
from instaui.internal.ui.unwrap_reactive import unwrap_reactive
from instaui.internal.ui.use_dark import use_dark
from instaui.internal.ui.use_language import use_language
from instaui.internal.ui.use_page_title import use_page_title
from instaui.internal.ui.vfor import VFor as vfor
from instaui.internal.ui.vfor_item import VForItem as TVForItem
from instaui.internal.ui.vif import VIf as vif
from instaui.internal.ui.web_computed import web_computed as computed
from instaui.internal.ui.web_event import event
from instaui.internal.ui.web_watch import watch
from instaui.version import __version__

from .web_launcher import WebLauncher as server
from .webview_launcher import WebViewLauncher as webview



from __future__ import annotations
from instaui.internal.pipeline.ast_build.base import ComponentBuilderProtocol
from instaui.internal.pipeline.ast_build.element_parser import ElementParser
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.pipeline.ast_build.match_parser import MatchParser
from instaui.internal.pipeline.ast_build.vfor_parser import VForParser
from instaui.internal.pipeline.ast_build.vif_parser import VIfParser
from instaui.internal.ui._scope import Scope
from instaui.internal.ui.match import Match
from instaui.internal.ui.renderable import Renderable
from instaui.internal.ui.element import Element
from instaui.internal.ast import core as ast_core
from instaui.internal.pipeline.ast_build.context import AstBuilderContext
from instaui.internal.ui.vfor import VFor
from instaui.internal.ui.vif import VIf
from instaui.internal.ui.components.content import Content


class RenderBuilder:
    def __init__(
        self, ctx: AstBuilderContext, component_builder: ComponentBuilderProtocol
    ) -> None:
        self.ctx = ctx
        self._expr_builder = ExprBuilder(ctx)
        self._element_parser = ElementParser(ctx, self)
        self._vfor_parser = VForParser(ctx, self)
        self._vif_parser = VIfParser(ctx, self)
        self._match_parser = MatchParser(ctx, self)
        self._component_builder = component_builder

    def build(self, renderable: Renderable) -> ast_core.Render:
        method = getattr(self, f"_build_{type(renderable).__name__}", None)
        if method:
            return method(renderable)

        if isinstance(renderable, Element):
            return self._build_Element(renderable)

        raise ValueError(f"Unsupported renderable type: {type(renderable)}")

    def _build_Element(self, element: Element) -> ast_core.Element:
        return self._element_parser.parse(element)

    def _build_VFor(self, vfor: VFor) -> ast_core.VFor:
        return self._vfor_parser.parse(vfor)

    def _build_Scope(self, scope: Scope) -> ast_core.ComponentRefRender:
        return ast_core.ComponentRefRender(
            ref=self._component_builder.build_component_ref(scope)
        )

    def _build_VIf(self, vif: VIf):
        return self._vif_parser.parse(vif)

    def _build_Match(self, match: Match) -> ast_core.MatchRender:
        return self._match_parser.parse(match)

    def _build_Content(self, content: Content) -> ast_core.ContentRender:
        return ast_core.ContentRender(
            self._expr_builder.build(content._content, "literal")
        )



from instaui.internal.ast.symbol import SymbolId

TOP_LEVEL_VAR_HINT = "_$tlv"
COMPONENT_HINT = "_$cp"
VAR_HINT = "_$v"


class VariableNameContext:
    def __init__(self) -> None:
        # symbol → js name
        self._symbols: dict[SymbolId, str] = {}

        # independent counters
        self._var_counter = 0
        self._component_counter = 0
        self._top_level_var_counter = 0

    # ---------- variables ----------
    def declare_var(self, symbol: SymbolId) -> str:
        if symbol in self._symbols:
            return self._symbols[symbol]

        name = f"{VAR_HINT}{self._var_counter}"
        self._var_counter += 1
        self._symbols[symbol] = name
        return name

    # ---------- components ----------
    def declare_component(self, symbol: SymbolId) -> str:
        if symbol in self._symbols:
            return self._symbols[symbol]

        name = f"{COMPONENT_HINT}{self._component_counter}"
        self._component_counter += 1
        self._symbols[symbol] = name
        return name

    def declare_top_level_var(self, symbol: SymbolId) -> str:
        if symbol in self._symbols:
            return self._symbols[symbol]

        name = f"{TOP_LEVEL_VAR_HINT}{self._top_level_var_counter}"
        self._top_level_var_counter += 1
        self._symbols[symbol] = name
        return name

    # ---------- identifiers ----------
    def resolve(self, symbol: SymbolId) -> str:
        return self._symbols[symbol]



from typing import Any, Optional
from instaui.internal.ui._expr.literal_expr import UILiteralExpr
from instaui.internal.ui.app_context import get_current_container
from instaui.internal.ui.bindable import mark_used
from instaui.internal.ui.renderable import Renderable
from instaui.internal.ui.container import Container


class Match(Renderable):
    def __init__(self, cond: Any):
        """
        Creates a conditional logic container that renders different UI blocks
        based on the value of the given reactive reference. Each case is checked
        sequentially, and the first match will be displayed. A default case can be
        defined to handle unmatched values.

        Args:
            cond (Any): A reactive reference or plain value whose
                state determines which case block should be rendered.

        Example:
        .. code-block:: python
            a = ui.state("")

            html.input().vmodel(a)

            with ui.match(a) as mt:
                with mt.case("page1"):
                    html.paragraph("in page1 case")

                with mt.case("page2"):
                    html.paragraph("in page2 case")

                with mt.default():
                    html.paragraph("in default case")
        """

        super().__init__()
        get_current_container().add_child(self)

        mark_used(cond)
        self._cond = cond
        self._cases: list[MatchCase] = []
        self._default_case: Optional[DefaultCase] = None

    def __enter__(self):
        return MatchWrapper(self)

    def __exit__(self, *_):
        pass


class MatchWrapper:
    def __init__(self, host: Match) -> None:
        self.__host = host

    def case(self, value: Any):
        case = MatchCase(value)
        self.__host._cases.append(case)
        return case

    def default(self):
        default_case = DefaultCase()
        self.__host._default_case = default_case
        return default_case


class MatchCase(Container):
    def __init__(self, value: Any):
        self._value = UILiteralExpr.try_parse(value)
        self._renderables: list[Renderable] = []

    def add_child(self, renderable: Renderable):
        self._renderables.append(renderable)


class DefaultCase(Container):
    def __init__(self):
        self._renderables: list[Renderable] = []

    def add_child(self, renderable: Renderable):
        self._renderables.append(renderable)



from __future__ import annotations
from enum import Enum
from typing import Any, Optional

from fastapi import FastAPI
from instaui.systems.dataclass_system import dataclass, field


@dataclass(frozen=True)
class WebUserRunOptions:
    host: str = "127.0.0.1"
    port: int = 8080
    reload: bool = True
    reload_dirs: str = "."
    reload_includes: str = "*.py"
    reload_excludes: str = ".*, .py[cod], .sw.*, ~*"
    log_level: str = "info"
    workers: Optional[int] = None
    uds: Optional[str] = None
    kwargs: dict = field(default_factory=dict)


@dataclass(frozen=True)
class WebUserRunWithOptions:
    app: FastAPI
    tags: Optional[list[str | Enum]] = None
    dependencies: Optional[list[Any]] = None
    responses: Optional[dict[int | str, dict[str, Any]]] = None



from typing import Optional
from instaui.internal.ui.element import Element
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier


class LazyRender(Element):
    def __init__(
        self,
        *,
        height: Optional[str] = None,
        destroy_on_leave: Optional[bool] = None,
        margin: Optional[str] = None,
        root: Optional[str] = None,
        disable: Optional[bool] = None,
    ):
        """
        Represents a lazy rendering container that defers rendering of its content until it becomes visible in the viewport.

        Args:
            height (Optional[str]): The CSS height of the container when content is not yet rendered.
                                    Defaults to "200px" if not specified.
            destroy_on_leave (Optional[bool]): Whether to destroy the rendered content when it leaves the viewport.
                                            Defaults to True.
            margin (Optional[str]): The margin around the root intersection observer area, specified as a CSS string
                                    (e.g., "100px 0px"). Defaults to "0px".
            root (Optional[str]): A CSS selector string specifying the root element for the IntersectionObserver.
                                If not provided, the browser viewport is used as the root.
            disable (Optional[bool]): If True, disables lazy rendering behavior and renders content immediately.
                                    Defaults to False.

        Example:
        .. code-block:: python
            with ui.lazy_render():
                ui.text("This content will be rendered only when it becomes visible in the viewport.")
        """
        super().__init__("lazy-render")

        self.props(
            {
                "height": height,
                "destroyOnLeave": destroy_on_leave,
                "margin": margin,
                "root": root,
                "disable": disable,
            }
        )

    def hidden_slot(self):
        """
        Provides access to the 'hidden' slot, which displays content while the main content is not yet rendered
        or has been destroyed due to leaving the viewport.

        Returns:
            The slot context manager for the 'hidden' slot.

        Example:
        .. code-block:: python
            with ui.lazy_render() as lr:
                ui.text("Main content")
            with lr.hidden_slot():
                ui.text("Placeholder shown before main content is rendered")
        """
        return self.add_slot("hidden")

    def on_visibility(
        self,
        handler: EventMixin,
        *,
        extends: Optional[list] = None,
        modifier: Optional[list[TEventModifier]] = None,
    ):
        """
        Registers an event handler that is triggered when the visibility state of the lazy-rendered content changes.

        Args:
            handler (EventMixin): The event handler function to call when visibility changes.
                                It receives a boolean indicating whether the content is now visible.
            extends (Optional[list]): Additional event data to extend the payload.
            modifier (Optional[list[TEventModifier]]): List of event modifiers (e.g., prevent_default, stop_propagation).

        Example:
        .. code-block:: python
            @ui.event(inputs=[ui.event_context.e()])
            def visibility_changed(visible: bool):
                print(f"Content is now {'visible' if visible else 'hidden'}")

            with ui.lazy_render().on_visibility(visibility_changed):
                ui.text("Content")
        """
        return self.on("visibility", handler, extends=extends, modifier=modifier)



from dataclasses import dataclass, field
from enum import Enum
from typing import List


class TemplateTagKind(Enum):
    HTML = "html"
    COMPONENT = "component"
    DYNAMIC = "dynamic"


@dataclass(frozen=True)
class TemplateDirectiveInfo:
    """
    Confirmed during normalize phase:
    - This is a Vue template supported directive
    - Its CallExpr structure is valid
    """

    name: str  # if / for / show
    arg_count: int


@dataclass
class TemplateElementInfo:
    """
    Vue SFC template codegen specific, strongly typed annotation
    """

    # tag related
    tag_kind: TemplateTagKind

    # Whether props / events have been validated as template-safe
    props_is_object_literal: bool = False
    events_is_object_literal: bool = False

    # Directive information (one-to-one correspondence with el.dirs)
    directives: List[TemplateDirectiveInfo] = field(default_factory=list)



from __future__ import annotations
from typing import Any, Callable, Iterable, Union
from typing_extensions import TypedDict
import webview.http as http
from webview.guilib import GUIType
from instaui.systems.dataclass_system import dataclass


@dataclass(frozen=True)
class WebViewUserRunOptions:
    webview_start_args: WebviewStartArgs


class WebviewStartArgs(TypedDict, total=False):
    func: Union[Callable[..., None], None]
    args: Union[Iterable[Any], None]
    localization: dict[str, str]
    gui: Union[GUIType, None]
    http_server: bool
    http_port: Union[int, None]
    user_agent: Union[str, None]
    private_mode: bool
    storage_path: Union[str, None]
    menu: list[Any]
    server: type[http.ServerType]  # type: ignore
    server_args: dict[Any, Any]
    ssl: bool
    icon: Union[str, None]



import inspect
from pathlib import Path
from typing import Callable, Tuple, get_type_hints


def get_fn_params_infos(fn: Callable) -> list[Tuple[str, type]]:
    """Get the parameter names and types of a function

    Args:
        fn (function): _description_

    Returns:
        [('a', int), ('b', str)]
    """
    signature = inspect.signature(fn)
    type_hints = get_type_hints(fn)

    return [
        (name, type_hints.get(name, inspect._empty))
        for name in signature.parameters.keys()
    ]


def is_last_param_args(fn):
    """
    Returns:
        bool: True if the last parameter of the function is `*args`, False otherwise.
    """
    sig = inspect.signature(fn)
    params = list(sig.parameters.values())

    if not params:
        return False

    last_param = params[-1]
    return last_param.kind == inspect.Parameter.VAR_POSITIONAL


def get_function_location_info(func) -> Tuple[str, int, str]:
    """

    Returns:
        Tuple[str, int, str]: A tuple containing the file path, start line number, and function name.
    """
    file_path = inspect.getfile(func)
    _, start_line = inspect.getsourcelines(func)

    return str(Path(file_path).resolve()), start_line, func.__qualname__



from instaui.systems.dataclass_system import dataclass


@dataclass(frozen=True)
class SymbolId:
    """
    The "name" at the logical level
    - Variables
    - Components
    - Module members
    """

    uid: int



from typing import TypeVar, cast
from pydantic import BaseModel

from instaui.internal.ui._expr.codec_expr import DataCodecExpr
from instaui.internal.ui.path_var import PathTrackerBindable
from instaui.internal.ui.to_value import ToValue
from instaui.protocol.ui.state import StateReadable
from instaui.debug.api_boundary import user_api

from .bindable import BindableMixin
from .ref_base import RefBase

_T = TypeVar("_T")


class StateModel(BaseModel):
    pass


@user_api
def state(value: _T, deep_compare: bool = False) -> _T:
    """
    Creates a reactive state wrapper that enables automatic UI updates when its value or
    nested elements change. Supports primitives, lists, dictionaries, and provides
    reactive operations such as indexing, comparisons, and length evaluation.

    Args:
        value (_T): The initial value to be wrapped as a reactive state. Can be any
            Python type including primitives, lists, or dictionaries.
        deep_compare (bool): Whether state updates should compare nested values to
            determine change propagation. Defaults to False.

    Example:
    .. code-block:: python
        from instaui import ui, html

        # Value reactivity
        s = ui.state('hello')
        html.input(s)
        ui.text(s)


        # List reactivity
        items = ui.state([1, 2, 3])
        html.number(items[0])
        html.ul.from_list(items)


        # Dictionary mutation triggers UI updates
        data = ui.state({"name": "John"})
        html.input(data["name"])


        # State used with js_computed
        data = ui.state({"items": [1, 2, 3]})
        items = ui.js_computed(inputs=[data], code="(d)=> d.items")
        html.ul.from_list(items)


        # Boolean operations maintain reactivity
        value = ui.state([{"state": True}])
        html.checkbox(value[0]["state"])
        html.paragraph(ui.not_(value[0]["state"]))


        # String or list length as reactive value
        text = ui.state("Hello")
        html.paragraph(ui.str_format("len: {}", ui.len_(text)))


        # Comparison operators trigger updates
        a = ui.state(1)
        b = ui.state(2)
        html.paragraph(a < b)  # UI updates when a or b changes
    """

    if isinstance(value, RefBase):
        return value

    if isinstance(value, StateReadable) or _readable_source_of_path_tracker(value):
        value = cast(_T, ToValue(value))

    new_value = value if isinstance(value, (BindableMixin)) else DataCodecExpr(value)

    return cast(_T, RefBase(value=new_value, deep_compare=deep_compare))


def _readable_source_of_path_tracker(value):
    return isinstance(value, PathTrackerBindable) and isinstance(
        value._source, StateReadable
    )



from abc import ABC, abstractmethod


class WebRuntimeServices(ABC):
    @abstractmethod
    def render_fn(self, route: str) -> str: ...

    @abstractmethod
    async def render_fn_async(self, route: str) -> str: ...



from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

from instaui import __version__ as _INSTA_VERSION
from instaui import resources
from instaui.constants.runtime import RuntimeMode
from instaui.internal.assets.context import GLOBAL_ASSETS, get_active_assets
from instaui.internal.assets.snapshot import AssetsSnapshot
from instaui.internal.ast.expression import RawFunctionExpr
from instaui.internal.backend.binding import BackendBindingRegistry
from instaui.internal.backend.webview.webview_backend import WebViewBackend
from instaui.internal.codegen.program_codegen import ProgramCodegen
from instaui.internal.pipeline.asset_resolver.webview_resolver import (
    WebViewAssetResolver,
)
from instaui.internal.pipeline.ast_build.builder import AstBuilder
from instaui.internal.pipeline.ast_build.context import AstBuildInput
from instaui.internal.pipeline.routing.fastapi import FastAPIRouteSyntaxAdapter
from instaui.internal.pipeline.webview_html.html_renderer import (
    HtmlOptions,
    WebViewHtmlRenderer,
)
from instaui.internal.router.registry import route_registry
from instaui.internal.runtime.app_meta import AppIcons, AppMeta, ServerRoute
from instaui.internal.runtime.assets_init import inject_framework_assets
from instaui.internal.runtime.layout import LayoutContext
from instaui.internal.runtime.render_ctx import RenderContext
from instaui.internal.runtime.runtime_ctx import RuntimeContext
from instaui.internal.runtime.webview.config import (
    PageBuildStrategy,
    WebViewUserLaunchConfig,
)
from instaui.internal.runtime.webview.options import WebViewUserRunOptions
from instaui.internal.ui._app import App
from instaui.internal.ui.app_context import get_default_app
from instaui.internal.ui.app_index import new_app

from .assets import FileSystemAssetMaterializer, PageAssetManager
from .assets_pipeline import WEBVIEW_ASSETS_PIPELINE

_INDEX_HTML_NAME = "index.html"


class WebViewRuntime:
    def __init__(self, config: WebViewUserLaunchConfig):
        self._launch_config = config
        self.backend = WebViewBackend(config)
        self.asset_materializer = FileSystemAssetMaterializer()

        self.route_syntax_adapter = FastAPIRouteSyntaxAdapter()
        self.binding_registry = BackendBindingRegistry(self.backend)
        self._mode = RuntimeMode.WEBVIEW
        self.debug = config.debug
        self._app_meta = AppMeta(self._mode, _INSTA_VERSION, config.debug)

        self.routes = route_registry.routes
        self.runtime_ctx = RuntimeContext(binding_registry=self.binding_registry)
        self.asset_manager = PageAssetManager(
            assets_root=self._launch_config.assets_path,
            clean_on_start=self._launch_config.clean_assets_on_start,
        )

        inject_framework_assets()

    def _render_page_core(self, route: str) -> str:
        page_dir = self.asset_manager.prepare_page_dir(route)
        layout_ctx = LayoutContext(get_default_app().layouts)

        with new_app(self._mode, debug=self.debug) as app:
            with layout_ctx:
                self.routes[route].fn()

            updated_app_meta = self._update_app_meta(app, route)
            page_assets = get_active_assets()
            assets_snapshot = WEBVIEW_ASSETS_PIPELINE.run(GLOBAL_ASSETS, page_assets)

        render_ctx = RenderContext(app_meta=updated_app_meta, route=route)

        asset_resolver = WebViewAssetResolver(page_dir, self.asset_materializer)

        ast = AstBuilder(self.runtime_ctx, render_ctx, asset_resolver).build(
            self._create_ast_build_input(app, assets_snapshot)
        )

        if self._launch_config.on_page_mounted:
            ast.components[0].on_mounted_calls.append(
                RawFunctionExpr("pywebview.api.on_app_mounted()")
            )

        code = ProgramCodegen(self.runtime_ctx).emit(ast)
        renderer = WebViewHtmlRenderer()
        html_str = renderer.render(
            code, self._build_render_options(assets_snapshot, asset_resolver)
        )

        index_html_path = page_dir / _INDEX_HTML_NAME
        index_html_path.write_text(html_str, encoding="utf-8")
        url = index_html_path.resolve().as_uri()

        self.asset_manager.mark_page_generated(route)

        return url

    def create_window(self, page_url: str = "/"):
        if not self.asset_manager.is_page_generated(page_url):
            url = self._render_page_core(page_url)
        else:
            page_dir = self._launch_config.assets_path.joinpath(page_url)
            url = (page_dir / _INDEX_HTML_NAME).resolve().as_uri()

        return self.backend.create_window(url)

    def run(self, optons: WebViewUserRunOptions):
        self._prebuild_pages()

        auto_create_window = self._launch_config.auto_create_window
        if auto_create_window:
            self.create_window(
                "/" if auto_create_window is True else auto_create_window
            )

        self.backend.run(optons)

    def _run_with_index(
        self, index_page_url: str, optons: WebViewUserRunOptions | None
    ):
        self.backend._run_with_index(index_page_url, optons)

    def _create_ast_build_input(
        self, app: App, assets_pool: AssetsSnapshot
    ) -> AstBuildInput:
        return AstBuildInput(app=app, scopes=app.created_scopes, assets=assets_pool)

    def _build_render_options(
        self, merged_assets: AssetsSnapshot, asset_resolver: WebViewAssetResolver
    ) -> HtmlOptions:
        import_maps = (
            merged_assets.import_maps
            # | self._options.get_import_maps_cdn_overrides()
        )

        return HtmlOptions(
            favicon_url=self._normalize_asset_path(
                merged_assets.favicon or resources.FAVICON_PATH, asset_resolver
            ),
            css_links=[
                self._normalize_asset_path(css, asset_resolver)
                for css in merged_assets.css_links
            ],
            js_links=[
                self._normalize_asset_path(js, asset_resolver)
                for js in merged_assets.js_links
            ],
            import_map={
                name: self._normalize_asset_path(link, asset_resolver)
                for name, link in import_maps.items()
            },
            style_tags=[
                (style_tag.group_id, style_tag.content)
                for style_tag in merged_assets.style_tags
            ],
            script_tags=[
                (script_tag.script_attrs, script_tag.content)
                for script_tag in merged_assets.script_tags
            ],
        )

    def _normalize_asset_path(
        self, path: Union[str, Path, None], asset_resolver: WebViewAssetResolver
    ) -> str:
        if path is None:
            return ""
        if isinstance(path, str):
            return path

        return asset_resolver.resolve_asset(path)

    def get_user_assets_folder(self, base_folder: Path) -> Optional[Path]:
        assets_folder = base_folder.joinpath("assets")
        if assets_folder.exists():
            return assets_folder

        return None

    def _update_app_meta(self, app: App, route: str):
        icons = AppIcons()
        if app._used_icons:
            icons.names = list(app._used_icons)
        if app._used_icons_set:
            icons.sets = list(app._used_icons_set)

        if not icons:
            icons = None

        return self._app_meta.replace(
            app_icons=icons,
            route=ServerRoute(self.route_syntax_adapter.normalize(route)),
        ).to_dict()

    def _is_param_route(self, route: str) -> bool:
        return "{" in route and "}" in route

    def _resolve_prebuild_routes(self) -> list[str]:
        strategy = self._launch_config.page_build_strategy
        prebuild = self._launch_config.prebuild_routes

        all_routes = list(self.routes.keys())

        if strategy == PageBuildStrategy.EAGER:
            routes_to_build = all_routes if prebuild is None else prebuild
        else:
            routes_to_build = prebuild or []

        # Check for parameterized routes
        param_routes = [r for r in routes_to_build if self._is_param_route(r)]
        unknown_routes = [r for r in routes_to_build if r not in self.routes]

        if param_routes or unknown_routes:
            msg_lines = ["Cannot prebuild the following routes in WebView mode:"]
            for r in param_routes:
                msg_lines.append(f"  - Parameterized route: {r}")
            for r in unknown_routes:
                msg_lines.append(f"  - Unknown route: {r}")
            msg_lines.append(
                "Prebuild routes must be fully static paths defined in @ui.page."
            )
            raise ValueError("\n".join(msg_lines))

        return routes_to_build

    def _prebuild_pages(self) -> None:
        routes_to_build = self._resolve_prebuild_routes()
        if not routes_to_build:
            return

        for route in routes_to_build:
            self._render_page_core(route)



from __future__ import annotations
import inspect
from typing import Any, Callable, Sequence
from pydantic import TypeAdapter, RootModel

from instaui.protocol.codec.type_adapter import TypeAdapterProtocol


JSONValue = Any  # dict | list | str | int | float | bool | None


class JsonCodec:
    def __init__(self) -> None:
        self._type_adapter_cache: dict[type, TypeAdapterProtocol] = {}

    def python_to_json_value(self, value: Any) -> JSONValue:
        return RootModel(value).model_dump()

    def build_type_adapter_map(
        self, func: Callable, inputs: Sequence
    ) -> dict[int, Callable[[Any], Any]] | None:
        if not inputs:
            return None

        sig = inspect.signature(func)
        params = list(sig.parameters.values())

        adapter_map: dict[int, Callable[[Any], Any]] = {}

        for idx, (param, _) in enumerate(zip(params, inputs)):
            annotation = param.annotation

            if annotation is inspect._empty:
                continue

            adapter_validator = self._get_type_adapter_validator(annotation)
            adapter_map[idx] = adapter_validator

        return adapter_map

    def _get_type_adapter_validator(self, tp: type) -> TypeAdapterProtocol:
        if tp not in self._type_adapter_cache:
            adapter = TypeAdapter(tp)
            self._type_adapter_cache[tp] = lambda value: adapter.validate_python(value)
        return self._type_adapter_cache[tp]


json_codec = JsonCodec()


def python_to_json_value(value: Any) -> JSONValue:
    return json_codec.python_to_json_value(value)


def build_type_adapter_map(
    func: Callable, inputs: Sequence
) -> dict[int, Callable[[Any], Any]] | None:
    return json_codec.build_type_adapter_map(func, inputs)


class JsonCodecError(TypeError):
    pass



from typing import Optional, Sequence
from instaui.internal.ui.protocol import CanOutputProtocol


def ensure_output_list(outputs: Optional[Sequence] = None):
    if outputs is None:
        return

    for output in outputs:
        if not isinstance(output, CanOutputProtocol):
            raise TypeError("The outputs parameter must be a `ui.state`")



from __future__ import annotations
import abc
from typing import TYPE_CHECKING, Optional
from instaui.internal.ui.event import EventMixin
from instaui.internal.ui.event_modifier import TEventModifier


if TYPE_CHECKING:
    from instaui.internal.ui.element import Element


class InputEventMixin:
    @abc.abstractmethod
    def _input_event_mixin_element(self) -> Element:
        pass

    def on_change(
        self,
        handler: EventMixin,
        *,
        extends: Optional[list] = None,
        modifier: Optional[list[TEventModifier]] = None,
    ):
        self._input_event_mixin_element().on(
            "change", handler, extends=extends, modifier=modifier
        )
        return self

    def on_input(
        self,
        handler: EventMixin,
        *,
        extends: Optional[list] = None,
        modifier: Optional[list[TEventModifier]] = None,
    ):
        self._input_event_mixin_element().on(
            "input", handler, extends=extends, modifier=modifier
        )
        return self



from instaui.internal.assets.style_assets import StyleTag


def gen_style_tag_group_id_attr(style_tag: StyleTag) -> str:
    if style_tag.group_id:
        return f' data-instaui-group-id="{style_tag.group_id}"'

    return ""



from __future__ import annotations
from pathlib import Path
from typing import Optional
from dataclasses import dataclass, field
from jinja2 import Environment, FileSystemLoader
from instaui.internal.codegen.program_codegen import CodegenResult
from .html_context import HtmlRenderContext


class WebViewHtmlRenderer:
    def __init__(self):
        self.env = Environment(
            loader=FileSystemLoader(Path(__file__).parent / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def render(self, codegen: CodegenResult, options: HtmlOptions) -> str:
        template = self.env.get_template("webview.html.j2")

        ctx = HtmlRenderContext(
            app_var=codegen.app_var,
            title=options.title,
            favicon_url=options.favicon_url,
            css_links=options.css_links,
            style_tags=options.style_tags,
            js_links=options.js_links,
            script_tags=options.script_tags,
            entry_code=codegen.entry_code,
            import_map=options.import_map if options.import_map else None,
            icons_svg_code=options.icons_svg_code,
        )

        return template.render(model=ctx)


@dataclass()
class HtmlOptions:
    favicon_url: str
    title: str = "instaui"

    css_links: list[str] = field(default_factory=list)
    style_tags: list[tuple[Optional[str], str]] = field(default_factory=list)
    """(group_id, content)"""
    js_links: list[str] = field(default_factory=list)

    #
    script_tags: list[tuple[Optional[dict], str]] = field(default_factory=list)
    """(attrs_dict, content)"""
    import_map: Optional[dict[str, str]] = None
    icons_svg_code: Optional[str] = None



import asyncio
from fastapi import Request, Depends, APIRouter
from fastapi.responses import StreamingResponse
import uuid

DEBUG_SSE_URL = "/instaui/debug-sse"

_task_events: dict[str, asyncio.Event] = {}


def create_router(router: APIRouter, debug_mode: bool):
    if debug_mode:
        _create_sse(router)


async def event_generator(
    request: Request, connection_id: str, interval_heart_beat_sec: float = 0.8
):
    task_event = asyncio.Event()
    _task_events[connection_id] = task_event

    try:
        while not task_event.is_set():
            if await request.is_disconnected():
                break

            yield "data:1\n\n"
            await asyncio.sleep(interval_heart_beat_sec)

    except asyncio.CancelledError:
        pass
    finally:
        if connection_id in _task_events:
            del _task_events[connection_id]


def _get_connection_id(request: Request):
    return str(uuid.uuid4())


def _create_sse(router: APIRouter):
    @router.get(DEBUG_SSE_URL)
    async def events(
        request: Request, connection_id: str = Depends(_get_connection_id)
    ):
        return StreamingResponse(
            event_generator(request, connection_id), media_type="text/event-stream"
        )


def when_server_reload():
    for task_id, task in _task_events.items():
        task.set()

    _task_events.clear()



from __future__ import annotations
from types import FrameType
import typing
import uvicorn
import socket
import logging

logger = logging.getLogger(__name__)
ThandleExitCallbacks = list[typing.Callable[[], typing.Any]]


class UvicornServer(uvicorn.Server):
    _instance: UvicornServer

    def __init__(self, config: uvicorn.Config) -> None:
        super().__init__(config)
        self._handle_exit_callbacks: ThandleExitCallbacks = []

    @classmethod
    def get_instance(cls) -> UvicornServer:
        if not hasattr(cls, "_instance"):
            raise RuntimeError("UvicornServer singleton not created")
        return cls._instance

    @classmethod
    def create_singleton(
        cls, config: uvicorn.Config, handle_exit_callbacks: ThandleExitCallbacks
    ) -> None:
        cls._instance = cls(config=config)
        for callback in handle_exit_callbacks:
            cls._instance.on_handle_exit(callback)

    def run(self, sockets: typing.Optional[list[socket.socket]] = None) -> None:
        self._instance = self
        super().run()

    def on_handle_exit(self, callback: typing.Callable[[], typing.Any]):
        self._handle_exit_callbacks.append(callback)

    def handle_exit(self, sig: int, frame: FrameType | None) -> None:
        for callback in self._handle_exit_callbacks:
            try:
                callback()
            except Exception as e:
                logger.exception("Exception in handle_exit callback: %s", e)
        return super().handle_exit(sig, frame)



from __future__ import annotations
from instaui.internal.ast.symbol import SymbolId


class ComponentFileMap:
    def __init__(self) -> None:
        self._map: dict[SymbolId, str] = {}

    def register(self, symbol: SymbolId, path: str) -> None:
        self._map[symbol] = path

    def resolve(self, symbol: SymbolId) -> str:
        return self._map[symbol]



from instaui.internal.assets.base import AssetsDeclaration
from instaui.internal.assets.css_assets import CssAsset


class PluginCssProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for plugin in assets.plugins:
            for css in plugin.css:
                # CSS resources with plugin namespaces should not override directly added CSS resources
                if (
                    isinstance(css, CssAsset)
                    and css.namespace is not None
                    and css.namespace in assets.css_links.by_namespace
                ):
                    continue

                assets.css_links.add(css)



from __future__ import annotations
import threading
from typing import TYPE_CHECKING, Callable, Hashable, Mapping, Optional
from fastapi import APIRouter
from instaui.internal.backend.common import handler as _handler
from instaui.protocol.codec.type_adapter import TypeAdapterProtocol
from . import _utils

if TYPE_CHECKING:
    from .endpoints import ServerEndpoints

_watch_handlers: dict[Hashable, _handler.HandlerInfo] = {}
dict_lock = threading.Lock()


def create_router(router: APIRouter, debug_mode: bool, endpoints: ServerEndpoints):
    _async_handler(router, endpoints)
    _sync_handler(router, endpoints)

    if debug_mode:

        @router.get("/instaui/watch-infos", tags=["instaui-debug"])
        def watch_infos():
            return get_statistics_info()


def _async_handler(router: APIRouter, endpoints: ServerEndpoints):
    @router.post(endpoints.ASYNC_WATCH_URL)
    async def _(data: dict):
        hkey = data.pop("hKey")
        handler_info = _watch_handlers.get(hkey)
        if handler_info is None:
            raise _utils.HandlerNotFoundError("watch handler not found")

        result = await handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return _utils.response_web_data(handler_info.outputs_binding_count, result)


def _sync_handler(router: APIRouter, endpoints: ServerEndpoints):
    @router.post(endpoints.WATCH_URL)
    def _(data: dict):
        hkey = data.pop("hKey")
        handler_info = _watch_handlers.get(hkey)
        if handler_info is None:
            raise _utils.HandlerNotFoundError("watch handler not found")

        result = handler_info.fn(
            *handler_info.get_handler_args(_get_binds_from_data(data))
        )
        return _utils.response_web_data(handler_info.outputs_binding_count, result)


def register_watch_handler(
    key: str,
    handler: Callable,
    outputs_binding_count: int,
    custom_type_adapter_map: Optional[Mapping[int, TypeAdapterProtocol]] = None,
):
    if key in _watch_handlers:
        return
    with dict_lock:
        _watch_handlers[key] = _handler.HandlerInfo.from_handler(
            handler, outputs_binding_count, custom_type_adapter_map
        )


def _get_binds_from_data(data: dict):
    return data.get("input", [])


def get_statistics_info():
    return {
        "_watch_handlers count": len(_watch_handlers),
        "_watch_handlers keys": list(_watch_handlers.keys()),
    }



from typing import Any
from instaui.systems import json_system
from instaui.internal import import_presets
from instaui.internal.ast.expression import (
    Expression,
    JsonLiteralExpr,
    Literal,
    IdentifierRef,
    MemberExpr,
    CallExpr,
    ObjectLiteral,
    ArrowFunctionExpr,
    ListLiteral,
    ObjectProperty,
    RawFunctionExpr,
    ObjectExpr,
    ListExpr,
    PathTrackerBindableCall,
    StrFormatCall,
    NullExpr,
    StringLiteral,
    UndefinedExpr,
    BackendValueRefExpr,
    ToValueCall,
)
from instaui.internal.ast.property_key import ExpressionKey, StringKey
from instaui.internal.codegen.context.file_ctx import FileCodegenContext


class ExpressionCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx

    def emit(self, expr: Expression) -> str:
        method = getattr(self, f"_emit_{type(expr).__name__}", None)
        if method:
            return method(expr)

        return self._emit_primitive(expr)

    def _emit_primitive(self, expr: Any) -> str:
        if isinstance(expr, bool):
            return "true" if expr else "false"

        if isinstance(expr, (str, int, float)):
            return repr(expr)

        if isinstance(expr, list):
            return "[" + ", ".join(self._emit_primitive(v) for v in expr) + "]"

        if isinstance(expr, dict):
            return (
                "{"
                + ", ".join(
                    f"{k}: {self._emit_primitive(v)}"
                    for k, v in expr.items()
                    if v is not None
                )
                + "}"
            )

        raise NotImplementedError(type(expr))

    def _emit_Literal(self, expr: Literal) -> str:
        return self.emit(expr.value)

    def _emit_IdentifierRef(self, expr: IdentifierRef) -> str:
        sid = expr.id
        return self.ctx.names.resolve(sid)

    def _emit_MemberExpr(self, expr: MemberExpr) -> str:
        obj = expr.obj
        member = expr.member
        return f"{self.emit(obj)}.{member}"

    def _emit_CallExpr(self, expr: CallExpr) -> str:
        callee = expr.callee
        args = expr.args
        args_code = ""
        if args:
            args_code = ", ".join(self.emit(a) for a in args)

        target = self.emit(callee)
        return f"{target}({args_code})"

    def _emit_ObjectLiteral(self, expr: ObjectLiteral) -> str:
        return self._emit_primitive(expr.props)

    def _emit_ArrowFunctionExpr(self, expr: ArrowFunctionExpr) -> str:
        body = expr.body
        return f"({self.emit(expr.params) if expr.params else ''}) => {self.emit(body) if body else r'{}'}"

    def _emit_ListLiteral(self, expr: ListLiteral) -> str:
        return self._emit_primitive(expr.value)

    def _emit_RawFunctionExpr(self, expr: RawFunctionExpr) -> str:
        return expr.js_code

    def _emit_ObjectExpr(self, expr: ObjectExpr) -> str:
        parts = []
        for prop in expr.props:
            parts.append(self._emit_property(prop))

        return "{" + ", ".join(parts) + "}"

    def _emit_ListExpr(self, expr: ListExpr) -> str:
        items = [self.emit(v) for v in expr.value]
        return "[" + ", ".join(items) + "]"

    def _emit_PathTrackerBindableCall(self, expr: PathTrackerBindableCall):
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.track_path())
        return f"{method}({self.emit(expr.target)},{self.emit(expr.args)})"

    def _emit_StrFormatCall(self, expr: StrFormatCall):
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.str_format())
        return f"{method}({self.emit(expr.args)})"

    def _emit_NullExpr(self, expr: NullExpr) -> str:
        return "null"

    def _emit_UndefinedExpr(self, expr: UndefinedExpr) -> str:
        return "undefined"

    def _emit_JsonLiteralExpr(self, expr: JsonLiteralExpr) -> str:
        return json_system.to_json_str(expr.value)

    def _emit_BackendValueRefExpr(self, expr: BackendValueRefExpr) -> str:
        return self._emit_primitive(
            self.ctx.root.binding_registry.get_value(expr.ref_id)
        )

    def _emit_StringLiteral(self, expr: StringLiteral) -> str:
        return json_system.to_json_str(
            self.emit(expr.value) if isinstance(expr.value, Expression) else expr.value
        )

    def _emit_ToValueCall(self, expr: ToValueCall) -> str:
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.to_value())
        return f"{method}({self.emit(expr.value)})"

    def _emit_property(self, prop: ObjectProperty) -> str:
        if isinstance(prop.key, str):
            key_str = prop.key

        if isinstance(prop.key, StringKey):
            key_str = f'"{prop.key.value}"'
        elif isinstance(prop.key, ExpressionKey):
            key_str = self.emit(prop.key.expr)

        if prop.shorthand:
            return key_str

        if prop.value is None:
            value_str = "undefined"
        else:
            value_str = self.emit(prop.value)

        return f"{key_str}: {value_str}"



from instaui.internal import import_presets
from instaui.internal.ast.core import (
    BackendComputedDecl,
    CustomRefDecl,
    ElementRefDecl,
    ExprComputedDecl,
    ExprEventDecl,
    JsComputedDecl,
    JsEventDecl,
    JsFnDecl,
    RefDecl,
    VariableDecl,
    BackendEventDecl,
    ConstDataDecl,
)
from instaui.internal.codegen.context.file_ctx import FileCodegenContext
from instaui.internal.codegen.exception import CodegenError
from .expr_codegen import ExpressionCodegen


class VariableDeclCodegen:
    def __init__(self, ctx: FileCodegenContext) -> None:
        self.ctx = ctx
        self.import_table = ctx.imports
        self.expr = ExpressionCodegen(ctx)

    def emit(self, var: VariableDecl) -> str:
        try:
            method = getattr(self, f"_emit_variable_{type(var).__name__}", None)
            if not method:
                raise NotImplementedError(type(var))
            return method(var)
        except Exception as e:
            raise CodegenError(f"Failed to generate code for {var}: {e}", var.source)

    def _emit_variable_RefDecl(self, var: RefDecl):
        method = self.import_table.use_from_preset(import_presets.Instaui.ref())
        target = self.ctx.names.resolve(var.target)
        args = self.expr.emit(var.args)
        return f"const {target} = {method}({args});"

    def _emit_variable_JsComputedDecl(self, var: JsComputedDecl):
        method = self.import_table.use_from_preset(import_presets.Instaui.js_computed())
        target = self.ctx.names.resolve(var.target)
        args = self.expr.emit(var.args)
        return f"const {target} = {method}({args});"

    def _emit_variable_BackendComputedDecl(self, var: BackendComputedDecl):
        method = self.import_table.use_from_preset(
            import_presets.Instaui.web_computed_ref()
        )
        target = self.ctx.names.resolve(var.target)

        computed_code = rf"const {target} = {method}({self.expr.emit(var.args)});"
        return computed_code

    def _emit_variable_JsEventDecl(self, var: JsEventDecl):
        method = self.import_table.use_from_preset(import_presets.Instaui.js_event())
        target = self.ctx.names.resolve(var.target)
        args = self.expr.emit(var.args)
        return f"const {target} = {method}({args});"

    def _emit_variable_BackendEventDecl(self, var: BackendEventDecl):
        method = self.import_table.use_from_preset(import_presets.Instaui.web_event())
        target = self.ctx.names.resolve(var.target)

        return rf"const {target} = {method}({self.expr.emit(var.args)});"

    def _emit_variable_ConstDataDecl(self, var: ConstDataDecl):
        target = self.ctx.names.resolve(var.target)
        value = self.expr.emit(var.value)
        return f"const {target} = {value};"

    def _emit_variable_ElementRefDecl(self, var: ElementRefDecl):
        method = self.ctx.imports.use_from_preset(import_presets.Vue.ref())
        target = self.ctx.names.resolve(var.target)
        return f"const {target} = {method}();"

    def _emit_variable_ExprComputedDecl(self, var: ExprComputedDecl):
        method = self.ctx.imports.use_from_preset(
            import_presets.Instaui.expr_computed()
        )
        target = self.ctx.names.resolve(var.target)
        return f"const {target} = {method}({self.expr.emit(var.args)});"

    def _emit_variable_ExprEventDecl(self, var: ExprEventDecl):
        method = self.ctx.imports.use_from_preset(import_presets.Instaui.expr_event())
        target = self.ctx.names.resolve(var.target)
        return f"const {target} = {method}({self.expr.emit(var.args)});"

    def _emit_variable_JsFnDecl(self, var: JsFnDecl):
        target = self.ctx.names.resolve(var.target)
        return f"const {target} = {self.expr.emit(var.fn)};"

    def _emit_variable_CustomRefDecl(self, var: CustomRefDecl):
        method = self.ctx.imports.use(module=var.module, name=var.name)
        target = self.ctx.names.resolve(var.target)
        args = ", ".join(self.expr.emit(p) for p in var.params) if var.params else ""
        return f"const {target} = {method}({args});"



from typing import Protocol
from instaui.internal.assets.base import AssetsDeclaration


class AssetsProcessor(Protocol):
    def process(self, assets: AssetsDeclaration) -> None:
        """In-place mutate assets"""
        ...



from pathlib import Path
from typing import Sequence
from instaui.constants.runtime import RuntimeMode
from instaui.internal.assets.base import AssetsDeclaration
from instaui.internal.assets.component_dep import ComponentDependencyInfo
from instaui.internal.assets.css_assets import CssAsset


class ComponentCssProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for comp in assets.component_dependencies.records:
            for css in comp.dependency.css:
                # CSS resources with component namespaces should not override directly added CSS resources
                if (
                    isinstance(css, CssAsset)
                    and css.namespace is not None
                    and css.namespace in assets.css_links.by_namespace
                ):
                    continue

                assets.css_links.add(css)


class ZeroModeComponentCssProcessor:
    def process(self, assets: AssetsDeclaration) -> None:
        for record in assets.component_dependencies.records:
            dependency = record.dependency
            css_resources = self._resolve_zero_mode_css(dependency)

            for css in css_resources:
                # CSS resources with component namespaces should not override directly added CSS resources
                if (
                    isinstance(css, CssAsset)
                    and css.namespace is not None
                    and css.namespace in assets.css_links.by_namespace
                ):
                    continue

                assets.css_links.add(css)

    @staticmethod
    def _resolve_zero_mode_css(
        dependency: ComponentDependencyInfo,
    ) -> Sequence[str | Path | CssAsset]:
        """
        CSS selection rules in Zero mode:

        - If `overrides[RuntimeMode.ZERO].zero_css` exists, **use this list exclusively**
        - Otherwise, fall back to `dependency.css`"
        """
        overrides = dependency.overrides
        if not overrides:
            return dependency.css

        zero_override = overrides.get(RuntimeMode.ZERO)
        if not zero_override or not zero_override.zero_css:
            return dependency.css

        return zero_override.zero_css



from __future__ import annotations
from typing import Optional
from instaui.internal.ast.property_key import StringKey
from instaui.internal.pipeline.ast_build.base import (
    RenderBuilderProtocol,
)
from instaui.internal.pipeline.ast_build.expr_builder import ExprBuilder

from instaui.internal.ui.event_args import EventArgs
from instaui.internal.ui.slot import Slot
from instaui.systems.dict_flow_system import (
    process_dict,
    remove_empty_collection,
    remove_none,
    repr_dict_keys,
)
from instaui.internal.ui.element import Element
from instaui.internal.ast import core as ast_core
from instaui.internal.ast import expression
from instaui.internal.pipeline.ast_build.context import AstBuilderContext


class ElementParser:
    def __init__(
        self, ctx: AstBuilderContext, render_builder: RenderBuilderProtocol
    ) -> None:
        self.ctx = ctx
        self._expr_builder = ExprBuilder(ctx)
        self._render_builder = render_builder

    def parse(self, element: Element) -> ast_core.Element:
        tag = self._expr_builder.build(element.tag)

        props = (
            process_dict(
                {
                    "bProps": element._binding_props,
                    "pProps": element._proxy_props,
                    "sProps": element._props,
                    "ref": element._element_ref,
                },
                [remove_none, remove_empty_collection, repr_dict_keys],
            )
            or None
        )

        props = self._create_props(element) if props else None
        events = self._create_events(element)
        dirs = self._create_directives(element)
        classes = self._create_classes(element)
        styles = self._create_styles(element)
        lifecycles = self._create_lifecycles(element)
        slots = self._create_slots(element)

        return ast_core.Element(
            tag=tag,
            props=props,
            events=events,
            dirs=dirs,
            classes=classes,
            styles=styles,
            lifecycles=lifecycles,
            slots=slots,
        )

    def _create_slots(self, element: Element) -> Optional[dict[str, ast_core.Slot]]:
        if not element._slots:
            return None

        return {
            name: self._parse_slot(name, slot) for name, slot in element._slots.items()
        }

    def _parse_slot(self, name: str, slot: Slot) -> ast_core.Slot:
        prop = None
        if slot._slot_prop._used:
            prop = self.ctx.var_mapper.get_or_create_id(slot._slot_prop)
            self.ctx.component_scope_manager.declare_var(prop)

        return ast_core.Slot(
            name=expression.StringLiteral(name),
            used_prop=prop,
            body=[self._render_builder.build(child) for child in slot.children],
        )

    def _create_lifecycles(
        self, element: Element
    ) -> Optional[dict[str, expression.Expression]]:
        if not element._lifecycle_events:
            return None

        return {
            event_name: expression.ListExpr(
                [self._event_call(event_arg) for event_arg in event_args]
            )
            for event_name, event_args in element._lifecycle_events.items()
        }

    def _create_styles(self, element: Element) -> Optional[ast_core.ElementStyles]:
        styles = ast_core.ElementStyles()
        if element._style:
            styles.static = expression.JsonLiteralExpr(element._style)

        if element._binging_style:
            styles.binding = expression.ObjectExpr.from_pairs(
                (StringKey(name), self._expr_builder.build(expr))
                for name, expr in element._binging_style.items()
            )

        if element._proxy_style:
            styles.proxy = [
                self._expr_builder.build(expr) for expr in element._proxy_style
            ]

        return styles or None

    def _create_classes(self, element: Element) -> Optional[ast_core.ElementClasses]:
        classes = ast_core.ElementClasses()
        if element._str_classes:
            classes.static = expression.ListLiteral(element._str_classes)

        if element._map_classes:
            classes.maps = expression.ObjectExpr.from_pairs(
                (StringKey(name), self._expr_builder.build(expr))
                for name, expr in element._map_classes.items()
            )

        if element._binding_classes:
            classes.binding = [
                self._expr_builder.build(expr) for expr in element._binding_classes
            ]

        return classes or None

    def _create_props(self, element: Element):
        props = ast_core.ElementProp()
        if element._binding_props:
            props.binding = {
                name: self._expr_builder.build(expr)
                for name, expr in element._binding_props.items()
            }
        if element._proxy_props:
            props.proxy = [
                self._expr_builder.build(expr) for expr in element._proxy_props
            ]

        if element._props:
            props.static = expression.JsonLiteralExpr(element._props)

        if element._element_ref:
            props.ref = self._expr_builder.build_Variable(element._element_ref)

        return props or None

    def _create_directives(
        self, element: Element
    ) -> Optional[list[ast_core.Directive]]:
        if not element._directives:
            return None

        return [
            ast_core.Directive(
                name=expression.Literal(directive.name),
                value=self._expr_builder.build(directive._value),
                arg=expression.Literal(directive._arg) if directive._arg else None,
                sys=expression.Literal(int(directive._is_sys))
                if directive._is_sys
                else None,
                mf=expression.ListLiteral(directive._modifiers)
                if directive._modifiers
                else None,
            )
            for directive in element._directives.keys()
        ]

    def _create_events(self, element: Element) -> Optional[expression.ObjectExpr]:
        if not element._events:
            return None

        return expression.ObjectExpr.from_pairs(
            (
                StringKey(event_name),
                expression.ListExpr(
                    [self._event_call(event_arg) for event_arg in event_args]
                ),
            )
            for event_name, event_args in element._events.items()
        )

    def _event_call(self, event_arg: EventArgs) -> expression.CallExpr:
        # _$v3.fn({extends: [], modifier: []})

        target = expression.IdentifierRef(
            self.ctx.var_mapper.get_or_create_id(event_arg.event)
        )

        args = {}
        if event_arg.extends:
            args["extends"] = self._expr_builder.list_expr(values=event_arg.extends)
        if event_arg.modifier:
            args["modifier"] = expression.ListLiteral(event_arg.modifier)

        return expression.CallExpr.of(
            expression.MemberExpr(obj=target, member="fn"),
            expression.ObjectExpr.from_kwargs(**args),
        )
